"""
    domonic.components
    ====================================

    A bit of a dumping ground for components ideas.

    Mainly used by the examples.

"""


import base64
import json

from domonic.html import *
from domonic.events import *


class Websocket(object):
    """[Creates a websocket with listeners for particular events]

    # TODO - collect other or all Window data and pass to a window object

    Args:
        str ([reference]): [the javascript variable name given to the socket]
        str ([address]): [where you want to connect to]
        str ([target]): [the dom element to attach the listeners to]
    """

    @staticmethod
    def get_event(msg):
        evt = None
        try:
            dom_event = json.loads(msg)
        except Exception as e:
            return  # pass on non json message

        # print(msg)
        event_string = dom_event['type']

        if event_string == "keydown" or event_string == "keyup":
            evt = KeyboardEvent(event_string)
            evt.keyCode = dom_event['keyCode']
            evt.charCode = dom_event['charCode']
            evt.code = dom_event['code']
            evt.key = dom_event['key']

        elif event_string == 'mousedown' or event_string == 'mouseup' or event_string == "mousemove":
            evt = MouseEvent(event_string)
            evt.initMouseEvent(_type=event_string, screenX=dom_event['screenX'], screenY=dom_event['screenY'],
                               clientX=dom_event['clientX'], clientY=dom_event['clientY'],
                               ctrlKey=dom_event['ctrlKey'], altKey=dom_event['altKey'], shiftKey=dom_event['shiftKey'])

        elif event_string in ["drag", "dragend", "dragenter", "dragexit", "dragleave", "dragover", "dragstart", "drop"]:
            evt = DragEvent(event_string)
            try:
                evt.dataTransfer = dom_event['dataTransfer']
            except Exception as e:
                pass

        elif event_string == "hashchange":
            # print(dom_event)
            evt = HashChangeEvent(event_string)
            evt.oldURL = dom_event['oldURL']
            evt.newURL = dom_event['newURL']

        elif event_string == 'cut' or event_string == 'copy' or event_string == "paste":
            # print("SUP::",event_string, dom_event)
            evt = ClipboardEvent(event_string)
            evt.clipboardData = dom_event['clipboardData']

        elif event_string == 'wheel':
            evt = WheelEvent(event_string)
            # print(evt)
            # print(dom_event)
            evt.deltaX = dom_event['deltaX']
            evt.deltaY = dom_event['deltaY']
            evt.deltaZ = dom_event['deltaZ']
            evt.deltaMode = dom_event['deltaMode']
            # ?? TODO - no deltaX? - myabe stripped by stringify? was on wrong target

        return evt

    def __init__(self, reference='socket', address='ws://0.0.0.0:5555', target="body",
                 mouse_events=True, keyboard_events=True,
                 ui_events=False, focus_events=False, touch_events=False, wheel_events=False,
                 animation_events=False, clipboard_events=False, error_events=False, submit_events=False,
                 pointer_events=False, before_unload_events=False, SVG_events=False, timer_events=False,
                 drag_events=False, hashchange_events=False, input_events=False, page_transition_events=False,
                 popstate_events=False, storage_events=False, transition_events=False, progress_events=False):

        self.reference = reference
        self.address = address
        self.mouse_events = mouse_events
        self.keyboard_events = keyboard_events
        self.ui_events = ui_events
        self.focus_events = focus_events
        self.touch_events = touch_events
        self.wheel_events = wheel_events
        self.animation_events = animation_events
        self.clipboard_events = clipboard_events
        self.error_events = error_events
        self.submit_events = submit_events
        self.pointer_events = pointer_events
        self.before_unload_events = before_unload_events
        self.SVG_events = SVG_events
        self.timer_events = timer_events
        self.drag_events = drag_events
        self.hashchange_events = hashchange_events
        self.input_events = input_events
        self.page_transition_events = page_transition_events
        self.popstate_events = popstate_events
        self.storage_events = storage_events
        self.transition_events = transition_events
        self.progress_events = progress_events

    def _add_listener(self, event, target='body'):
        return str('''
            $("''' + target + '''").on("''' + event + '''", function(event){
                socket.send( stringify_object(event) );
            });
            ''')
        # TODO - no jquery. detect targets and use # addEventListener

    def __str__(self):
        events = ''
        if self.mouse_events:
            events += self._add_listener('mousedown')
            events += self._add_listener('mousemove')
            events += self._add_listener('mouseup')
        if self.keyboard_events:
            events += self._add_listener('keydown')
            events += self._add_listener('keyup')
        if self.ui_events:  # https://www.w3schools.com/jsref/obj_uievent.asp
            events += self._add_listener('abort')
            events += self._add_listener('beforeunload')
            events += self._add_listener('error')
            events += self._add_listener('load')
            events += self._add_listener('resize')
            events += self._add_listener('scroll')
            events += self._add_listener('select')
            events += self._add_listener('unload')
        if self.focus_events:
            events += self._add_listener('blur')
            events += self._add_listener('focus')
            events += self._add_listener('focusin')
            events += self._add_listener('focusout')
        if self.touch_events:
            events += self._add_listener('touchstart')
            events += self._add_listener('touchend')
            events += self._add_listener('touchmove')
            events += self._add_listener('touchcancel')
        if self.wheel_events:
            # events += self._add_listener('wheel')
            events += '''
            window.addEventListener('wheel', function(){
                socket.send( stringify_object(event) );
            }, false);
            '''
        if self.animation_events:
            events += self._add_listener('animationend')
            events += self._add_listener('animationiteration')
            events += self._add_listener('animationstart')
        if self.clipboard_events:
            # events += self._add_listener('copy')
            # events += self._add_listener('cut')
            # events += self._add_listener('paste')
            events += '''
            window.addEventListener('cut', function(){
                socket.send( stringify_object(event) );
            }, false);
            window.addEventListener('copy', function(){
                socket.send( stringify_object(event) );
            }, false);
            window.addEventListener('paste', function(){
                socket.send( stringify_object(event) );
            }, false);
            '''
        # if self.error_events:
        # events += self._add_listener('')
        if self.submit_events:
            events += self._add_listener('submit')
        # if self.pointer_events:
            # events += '''
            # $("body").on("keydown", function(event){
            #     socket.send( stringify_object(event) );
            # })
            # $("body").on("keyup", function(event){
            #     socket.send( stringify_object(event) );
            # })
            # '''
        # if self.before_unload_events:
            # events += '''
            # $("body").on("keydown", function(event){
            #     socket.send( stringify_object(event) );
            # })
            # $("body").on("keyup", function(event){
            #     socket.send( stringify_object(event) );
            # })
            # '''
        # if self.SVG_events:
            # events += '''
            # $("body").on("keydown", function(event){
            #     socket.send( stringify_object(event) );
            # })
            # $("body").on("keyup", function(event){
            #     socket.send( stringify_object(event) );
            # })
            # '''
        # if self.timer_events:
            # events += '''
            # $("body").on("keydown", function(event){
            #     socket.send( stringify_object(event) );
            # })
            # $("body").on("keyup", function(event){
            #     socket.send( stringify_object(event) );
            # })
            # '''
        if self.drag_events:
            events += self._add_listener('drag')
            events += self._add_listener('dragend')
            events += self._add_listener('dragenter')
            events += self._add_listener('dragleave')
            events += self._add_listener('dragover')
            events += self._add_listener('dragstart')
            events += self._add_listener('drop')
        if self.hashchange_events:
            # events += self._add_listener('hashchange', "window")
            events += '''
            window.addEventListener('hashchange', function(){
                socket.send( stringify_object(event) );
            }, false);
            '''
        if self.input_events:
            events += self._add_listener('input')
        if self.page_transition_events:
            events += self._add_listener('pagehide')
            events += self._add_listener('pageshow')
        if self.popstate_events:
            events += self._add_listener('popstate')
        if self.storage_events:
            events += self._add_listener('storage')
        if self.transition_events:
            events += self._add_listener('transitionend')
        if self.progress_events:
            events += self._add_listener('error')
            events += self._add_listener('loadstart')

        return str(
            script('''
            const socket = new WebSocket("''' + self.address + '''");

            function stringify_object(object, depth=0, max_depth=2) {
                //console.log(object);
                // change max_depth to see more levels, for a touch event, 2 is good
                if (depth > max_depth)
                    return 'Object';

                const obj = {};
                for (let key in object) {
                    let value = object[key];
                    if (value instanceof Node)
                        // specify which properties you want to see from the node
                        value = {id: value.id};
                    else if (value instanceof Window)
                        value = 'Window';
                    else if (value instanceof Object)
                        value = stringify_object(value, depth+1, max_depth);

                    if(key=="originalEvent"){ // note im stripping this one off
                        continue;
                    }
                    obj[key] = value;
                }
                return depth? obj: JSON.stringify(obj);
            }

            $(document).ready(function(){
            ''' + events + """
            });"""
            )
        )


# class Video():
# class Sound():
# def JS(self, thing):
#     	if '.js' in thing[0:15]:
# 		return script( _src=thing )
# 	return script(thing)
# def CSS(self, thing):
# 	if '/' in thing[0:15]:
# 		return link( _rel="stylesheet", _href=f"{thing}" )
# 	return script(thing)


class SpriteCSS(object):
    """ a css sprite sheet.

    this spritesheet works by shifting the position of a bg image
    uses css animation. usage:

    animated_monster = SpriteCSS('monster', 190, 240, 'assets/spritesheets/monster.png', 0.8, 10)

    Args:
        _id - requires an ID which currently becomes a 'css class'
        width
        height
        spritesheet
        time
        steps
        loop
        y_offset - for spritesheets with mulitple rows you can offset the ypos
        bg_color

    Returns:
        str: A HTML rendered string

    """
    STYLE = lambda _id, width, height, spritesheet, time, steps, loop, y_offset, bg_color: """
        .""" + _id + """ {
          background:""" + bg_color + """;
          width:""" + str(width) + """px;
          height:""" + str(height) + """px;
          background: url('""" + spritesheet + """') left center;
          animation:""" + _id + """ """ + str(time) + """s steps(""" + str(steps) + """) infinite;
        }
        /*
        @keyframes """ + _id + """ {
            100% { background-position: -""" + str(steps * width) + """px; }
        }
        */
        @keyframes """ + _id + """ {
            from { background-position:0px -""" + str(y_offset) + """px; }
            to { background-position:-""" + str(steps * width) + """px -""" + str(y_offset) + """px; }
        }
    """

    def __init__(self, id, width, height, spritesheet, time, steps, loop=True, y_offset=0, bg_color="none"):
        self.id = id
        self.styles = SpriteCSS.STYLE(id, width, height, spritesheet, time, steps, loop, y_offset, bg_color)

    def __str__(self):
        return str(
            span(
                style(self.styles),
                div(_class=self.id)
            )
        )
# _ss = SpriteCSS( path )


# other sprite types may be needed...
# _ss = Sprite()
# _ss = BaseSprite()
# _ss = SpriteJS( path )
# _ss = SpriteSVG( path )
# _ss = SpriteGIF( path )
# _ss = SpriteShader( path )
# _ss = TileSet( path )


class DomonicJS(object):
    def __init__(self):
        pass

    def __str__(self):
        return script("""
            window.domonic = {'version':'0.0.2'}; // || domonic.js
            function print(msg){
                console.log(msg);
            }
            function title(str){
                return str.charAt(0).toUpper();
            }
            //class list() {
            //    constructor(list) {
            //        this.list = list;
            //    }
            //    append(item) {
            //        this.list.push(item);
            //    }
            //}
            """
        )

# class ImgButton():

#     def __init__(self, up, over=None, down=None, label=None, label_position=None):
#         self.up = up
#         self.over = over
#         self.down = down


# class ImgCheckbox():

#     def __init__(self, up, over=None, down=None, label=None, label_position=None):
#         self.up = up
#         self.over = over
#         self.down = down


class Sound(object):

    def __init__(self, filename):
        self.filename = filename

    def __str__(self):
        return str(
            div(
                audio(source(_src=self.filename, _type="audio/mp3"), _id="sfx", _autoplay="true"),
                script("""
                function play_sound( filename ) {
                let sound = document.getElementById("sfx");
                sound.currentTime = 0;
                sound.src = filename
                sound.play();
                };
                """, _type="text/javascript"
                ),
                _id='sound'
            )
        )


class ProgressBar(object):

    def __init__(self,):
        """
        a progress bar for loading or health / mana etc
        """
        # self.direction # the fill direction for the progress bar.
        # self.max
        # self.min
        self.percent
        self.value
        # self.style =

    def __str__(self):
        # $('#progress_bar span').css('width', somevalue);
        return str(
            div(span(_style="width:100%"), _id="progress_bar", _class="meter")
        )


class Input(object):

    BUTTON = "button"
    CHECKBOX = "checkbox"
    COLOR = "color"
    DATE = "date"
    DATETIME = "datetime-local"
    EMAIL = "email"
    FILE = "file"
    HIDDEN = "hidden"
    IMAGE = "image"
    MONTH = "month"
    NUMBER = "number"
    PASSWORD = "password"
    RADIO = "radio"
    RANGE = "range"
    RESET = "reset"
    SEARCH = "search"
    SUBMIT = "submit"
    TEL = "tel"
    TEXT = "text"
    TIME = "time"
    URL = "url"
    WEEK = "week"

    def __init__(self, _type, _id=None, _name=None, _label=None, *args, **kwargs):
        self._type = _type
        self._label = _label
        self._id = _id
        self._name = _name

    def __str__(self):
        # lab = str(label(self._label, _for=self._name))
        lab = str(label(str(self._label)))
        inp = str(input(_type=self._type, _id=self._id, _name=self._name))
        return str(lab + inp)


# WARNING. What is not documented is subject to lots of change!
# below are just examples of how to build your own components
# they may be removed in future version.
# You should build your own components libaries using domonic


class Modal(object):  # TODO - shouldn't this extend dom?

    def __init__(self, reference=None, content=None):
        self.reference = reference
        self.content = content

    def __str__(self):
        return str(
            div(
                div(
                    span("&times;", _class="close", **{"_data-ref": self.reference}),
                    div(self.content),
                    _class="modal-content",
                    _style="background-color:#fefefe;margin:15% auto;padding:20px;border:1px solid;width:80%;"
                ),
                _class="modal",
                _style="display:none;position:fixed;z-index:1;left:0;top:0;width:100%;height:100%; \
                    overflow:auto;background-color:rgb(0,0,0);background-color:rgba(0,0,0,0.4);",
                _id=self.reference
            )
        )


class Webpage(object):  # TODO - shouldn't this extend html?

    def __init__(self, content=None):
        self.content = content

    def __str__(self):
        classless_css = link(_rel="stylesheet", _href="https://unpkg.com/marx-css/css/marx.min.css")
        jquery = script(_src="https://code.jquery.com/jquery-3.5.1.min.js")
        code = script('''
            $(document).on( "click", ".close", function() {
                var _id = $(this).data('ref');
                $('#'+_id).css("display","none");
            });
            $(document).on( "click", ".open", function() {
                var _id = $(this).data('ref');
                $('#'+_id).css("display","block");
            });

            // When the user clicks anywhere outside of the modal, close it
            //window.onclick = function(event) {
            //    if (event.targeevent_string == modal) {
            //        modal.style.display = "none";
            //    }
            //}

            // pass an ElementID and an endpoint to redraw that div with the endpoints response
            window.redraw = function( _id, endpoint ){
                $.get( endpoint, function( data ) {
                window.console.log(data)
                $( "#"+_id ).html( $(data).html() );
                });
            }

        ''')
        styles = style('''
            .domonic-container {
                padding:20px;
            }
            .modal {
                display: none;
                position: fixed;
                z-index: 1;
                left: 0;
                top: 0;
                width: 100%;
                height: 100%;
                overflow: auto;
                background-color: rgb(0,0,0);
                background-color: rgba(0,0,0,0.4);
            }
            .modal-content {
                background-color: #fefefe;
                margin: 15% auto;
                padding: 20px;
                border: 1px solid #888;
                width: 80%;
            }
            .btn-sm {
                font-size:10px;
                padding: 0px;
                padding-left: 2px;
                padding-right: 2px;
            }
            .del {
                background-color:red;
            }
            .go {
                background-color:green;
            }

        ''')
        return str(
            html(
                '<!DOCTYPE HTML>',
                head(classless_css, jquery, code, styles),
                body(div(self.content, _class="domonic-container"))
            )
        )


# quick templates


webpage_tmpl = lambda content: f"""
    <!DOCTYPE html>
    <html>
    <head>
    <meta charset="utf-8">
    <title>webpage</title>
    </head>
    <body>
    {content}
    </body>
    </html>
    """

# carousel_tmpl = lambda content: f"""
#     <div class="carousel slide" data-ride="carousel">
#         <div class="carousel-inner">
#             {content}
#         </div>
#         <a class="left carousel-control" href="#myCarousel" role="button" data-slide="prev">
#             <span class="glyphicon glyphicon-chevron-left" aria-hidden="true"></span>
#             <span class="sr-only">Previous</span>
#         </a>
#         <a class="right carousel-control" href="#myCarousel" role="button" data-slide="next">
#             <span class="glyphicon glyphicon-chevron-right" aria-hidden="true"></span>
#             <span class="sr-only">Next</span>
#         </a>
#     </div>
#     """

# item_tmpl = lambda content: f"""
#     <div class="item">
#         {content}
#     </div>
#     """



"""
    domonic.webapi.history
    ====================================
    https://developer.mozilla.org/en-US/docs/Web/API/History
"""

# class History():
    # def __init__():
        # pass
    # def back():
    #     """ Loads the previous URL in the history list """
    #     raise NotImplementedError
    # def forward():
    #     """ Loads the next URL in the history list """
    #     raise NotImplementedError
    # def go():
    #     """ Loads a specific URL from the history list """
    #     raise NotImplementedError


# from domonic.html import *
# from domonic.js import *
# from domonic.css import *
# from domonic.events import *
# from domonic.internal import _get_window
# from domonic.internal import _get_document
# from domonic.internal import _get_location
# from domonic.internal import _get_history
# from domonic.internal import _get_navigator
# from domonic.internal import _get_screen
# from domonic.internal import _get_screen_left
# from domonic.internal import _get_screen_top
# from domonic.internal import _get_screen_width
# from domonic.internal import _get_screen_height
# from domonic.internal import _get_screen_avail_width
# from domonic.internal import _get_screen_avail_height
# from domonic.internal import _get_screen_pixel_ratio

# nice ideas from copilot

# but it also cheats... this will not work

# class History:

#     def __init__(self, window):
#         self._window = window

#     def back(self):
#         self._window.history.back()

#     def forward(self):
#         self._window.history.forward()

#     def go(self, steps):
#         self._window.history.go(steps)

#     def pushState(self, data, title, url):
#         self._window.history.pushState(data, title, url)

#     def replaceState(self, data, title, url):
#         self._window.history.replaceState(data, title, url)

#     def state(self):
#         return self._window.history.state()

#     def length(self):
#         return self._window.history.length()


"""
    test_geom
    ~~~~~~~~~~~~
"""

import unittest
# import requests
# from mock import patch

# from domonic import domonic
from domonic.geom import vec2, vec3, Shape, Rect, Line, Circle
from domonic.constants.color import Color


class TestCase(unittest.TestCase):

    # @silence
    def test_vec2(self):
        v = vec2(10, 10)
        assert v == v
        print(v * 2)
        print(v / 2)
        print(v - 2)
        print(v + v)
        print(v * v)
        print(v / v)
        print(v - v)
        print(v[0], v[1])
        print(v['x'], v['y'])
        print(v.x, v.y)

    # @silence
    def test_vec3(self):
        v = vec3(10, 10, 10)
        print(v * 2)
        print(v / 2)
        print(v - 2)
        print(v + v)
        print(v * v)
        # print(v / v)
        print(v - v)

    # @silence
    # def test_vec4(self):
    #     v = vec4(10, 10, 10)
    #     print(v)

    # @silence
    def test_shape(self):
        s = Shape(Color.red)
        print(s.color)

    # @silence
    def test_rect(self):
        r = Rect(0, 0, 10, 10)
        print(r)
        print(r * 10)
        print(r / 10)

        a = Rect(0, 0, 10, 10)
        b = Rect(0, 0, 20, 20)
        c = Rect(0, 0, 30, 30)
        print(a + b + c)

    # @silence
    def test_line(self):
        l = Line(vec2(0, 0), vec2(10, 10))
        print(l)

    # @silence
    def test_circle(self):
        c = Circle(0, 0, 10)
        print(c)
        print(c * 4)

    # @silence
    # def test_plotter(self):
        # p = Plotter()
        # print(p)

    # @silence
    def test_color(self):

        from domonic.constants.color import Color

        c1 = Color('#ff00ff')
        print(c1)
        # c2 = Color('red')
        # print(c2)
        c3 = Color(255, 255, 255)
        print(c3)
        c4 = Color(255, 255, 255, 0)
        print(c4)


if __name__ == '__main__':
    unittest.main()



"""
    domonic.constants.entities
    ====================================

"""


class Entity():

    # __slots__ = ('entity',)

    def __init__(self, entity: str):
        self.entity = entity

    def __str__(self):
        import html
        return html.unescape(self.character)


class Char():

    def __init__(self, character: str):
        self.character = character

    def __str__(self):
        import html
        return html.escape(self.character)

    # def __repr__(self):
    #     return self.character

    # web
    # ASCII Characters (Printable)
    SPACE = '&#32;'
    EXCLAMATION_MARK = '&#33;'  #: !
    QUOTATION_MARK = '&#34;'  #: "
    NUMBER_SIGN = '&#35;'  #: #
    DOLLAR_SIGN = '&#36;'  #: $
    PERCENT_SIGN = '&#37;'  #: %
    AMPERSAND = '&amp;'  #: &
    APOSTROPHE = '&#39;'  #: '
    OPENING_PARENTHESIS = '&#40;'  #: (
    LEFT_PARENTHESIS = '&#40;'  #: (
    CLOSING_PARENTHESIS = '&#41;'  #: )
    RIGHT_PARENTHESIS = '&#41;'  #: )
    ASTERISK = '&#42;'  #: *
    PLUS_SIGN = '&#43;'  #: +
    COMMA = '&#44;'  #: ,
    HYPHEN = '&#45;'  #: -
    PERIOD = '&#46;'  #: .
    SLASH = '&#47;'  #: /
    ZERO = '&#48;'  #: 0
    ONE = '&#49;'  #: 1
    TWO = '&#50;'  #: 2
    THREE = '&#51;'  #: 3
    FOUR = '&#52;'  #: 4
    FIVE = '&#53;'  #: 5
    SIX = '&#54;'  #: 6
    SEVEN = '&#55;'  #: 7
    EIGHT = '&#56;'  #: 8
    NINE = '&#57;'  #: 9

    COLON = '&#58;'  #: :
    SEMICOLON = '&#59;'  #: ;
    LESS_THAN = '&lt;'  #: <
    EQUALS_SIGN = '&#61;'  #: =
    GREATER_THAN = '&gt;'  #: >
    QUESTION_MARK = '&#63;'  #: ?
    AT_SIGN = '&#64;'  #: @

    UPPERCASE_A = '&#65;'  #: A
    UPPERCASE_B = '&#66;'  #: B
    UPPERCASE_C = '&#67;'  #: C
    UPPERCASE_D = '&#68;'  #: D
    UPPERCASE_E = '&#69;'  #: E
    UPPERCASE_F = '&#70;'  #: F
    UPPERCASE_G = '&#71;'  #: G
    UPPERCASE_H = '&#72;'  #: H
    UPPERCASE_I = '&#73;'  #: I
    UPPERCASE_J = '&#74;'  #: J
    UPPERCASE_K = '&#75;'  #: K
    UPPERCASE_L = '&#76;'  #: L
    UPPERCASE_M = '&#77;'  #: M
    UPPERCASE_N = '&#78;'  #: N
    UPPERCASE_O = '&#79;'  #: O
    UPPERCASE_P = '&#80;'  #: P
    UPPERCASE_Q = '&#81;'  #: Q
    UPPERCASE_R = '&#82;'  #: R
    UPPERCASE_S = '&#83;'  #: S
    UPPERCASE_T = '&#84;'  #: T
    UPPERCASE_U = '&#85;'  #: U
    UPPERCASE_V = '&#86;'  #: V
    UPPERCASE_W = '&#87;'  #: W
    UPPERCASE_X = '&#88;'  #: X
    UPPERCASE_Y = '&#89;'  #: Y
    UPPERCASE_Z = '&#90;'  #: Z

    OPENING_SQUARE_BRACKET = '&#91;'  #: [
    BACKSLASH = '&#92;'  #: \
    CLOSING_SQUARE_BRACKET = '&#93;'  #: ]
    CARET = '&#94;'  #: ^
    UNDERSCORE = '&#95;'  #: _
    GRAVE_ACCENT = '&#96;'  #:
    LOWERCASE_A = '&#97;'  #: a
    LOWERCASE_B = '&#98;'  #: b
    LOWERCASE_C = '&#99;'  #: c
    LOWERCASE_D = '&#100;'  #: d
    LOWERCASE_E = '&#101;'  #: e
    LOWERCASE_F = '&#102;'  #: f
    LOWERCASE_G = '&#103;'  #: g
    LOWERCASE_H = '&#104;'  #: h
    LOWERCASE_I = '&#105;'  #: i
    LOWERCASE_J = '&#106;'  #: j
    LOWERCASE_K = '&#107;'  #: k
    LOWERCASE_L = '&#108;'  #: l
    LOWERCASE_M = '&#109;'  #: m
    LOWERCASE_N = '&#110;'  #: n
    LOWERCASE_O = '&#111;'  #: o
    LOWERCASE_P = '&#112;'  #: p
    LOWERCASE_Q = '&#113;'  #: q
    LOWERCASE_R = '&#114;'  #: r
    LOWERCASE_S = '&#115;'  #: s
    LOWERCASE_T = '&#116;'  #: t
    LOWERCASE_U = '&#117;'  #: u
    LOWERCASE_V = '&#118;'  #: v
    LOWERCASE_W = '&#119;'  #: w
    LOWERCASE_X = '&#120;'  #: x
    LOWERCASE_Y = '&#121;'  #: y
    LOWERCASE_Z = '&#122;'  #: z

    OPENING_CURLY_BRACE = '&#123;'  #: {
    LEFT_CURLY_BRACE = '&#123;'  #: {
    VERTICAL_BAR = '&#124;'  #: |
    CLOSING_CURLY_BRACE = '&#125;'  #: }
    RIGHT_CURLY_BRACE = '&#125;'  #: }
    TILDE = '&#126;'  #: ~

    # ISO-8859-1 Characters
    AGRAVE = '&Agrave;'  #: À
    AACUTE = '&Aacute;'  #: Á
    ACIRC = '&Acirc;'  #: Â
    ATILDE = '&Atilde;'  #: Ã
    AUML = '&Auml;'  #: Ä
    ARING = '&Aring;'  #: Å
    AELIG = '&AElig;'  #: Æ
    CCEDIL = '&Ccedil;'  #: Ç
    EGRAVE = '&Egrave;'  #: È
    EACUTE = '&Eacute;'  #: É
    ECIRC = '&Ecirc;'  #: Ê
    EUML = '&Euml;'  #: Ë
    IGRAVE = '&Igrave;'  #: Ì
    IACUTE = '&Iacute;'  #: Í
    ICIRC = '&Icirc;'  #: Î
    IUML = '&Iuml;'  #: Ï
    ETH = '&ETH;'  #: Ð
    NTILDE = '&Ntilde;'  #: Ñ
    OGRAVE = '&Ograve;'  #: Ò
    OACUTE = '&Oacute;'  #: Ó
    OCIRC = '&Ocirc;'  #: Ô
    OTILDE = '&Otilde;'  #: Õ
    OUML = '&Ouml;'  #: Ö
    OSLASH = '&Oslash;'  #: Ø
    UGRAVE = '&Ugrave;'  #: Ù
    UACUTE = '&Uacute;'  #: Ú
    UCIRC = '&Ucirc;'  #: Û
    UUML = '&Uuml;'  #: Ü
    YACUTE = '&Yacute;'  #: Ý
    THORN = '&THORN;'  #: Þ
    SZLIG = '&szlig;'  #: ß
    AGRAVE = '&agrave;'  #: à
    AACUTE = '&aacute;'  #: á
    ACIRC = '&acirc;'  #: â
    ATILDE = '&atilde;'  #: ã
    AUML = '&auml;'  #: ä
    ARING = '&aring;'  #: å
    AELIG = '&aelig;'  #: æ
    CCEDIL = '&ccedil;'  #: ç
    EGRAVE = '&egrave;'  #: è
    EACUTE = '&eacute;'  #: é
    ECIRC = '&ecirc;'  #: ê
    EUML = '&euml;'  #: ë
    IGRAVE = '&igrave;'  #: ì
    IACUTE = '&iacute;'  #: í
    ICIRC = '&icirc;'  #: î
    IUML = '&iuml;'  #: ï
    ETH = '&eth;'  #: ð
    NTILDE = '&ntilde;'  #: ñ
    OGRAVE = '&ograve;'  #: ò
    OACUTE = '&oacute;'  #: ó
    OCIRC = '&ocirc;'  #: ô
    OTILDE = '&otilde;'  #: õ
    OUML = '&ouml;'  #: ö
    OSLASH = '&oslash;'  #: ø
    UGRAVE = '&ugrave;'  #: ù
    UACUTE = '&uacute;'  #: ú
    UCIRC = '&ucirc;'  #: û
    UUML = '&uuml;'  #: ü
    YACUTE = '&yacute;'  #: ý
    THORN = '&thorn;'  #: þ
    YUML = '&yuml;'  #: ÿ

    # ISO-8859-1 Symbols
    NBSP = '&nbsp;'  #:
    IEXCL = '&iexcl;'  #: ¡
    CENT = '&cent;'  #: ¢
    POUND = '&pound;'  #: £
    CURREN = '&curren;'  #: ¤
    YEN = '&yen;'  #: ¥
    BRVBAR = '&brvbar;'  #: ¦
    SECT = '&sect;'  #: §
    UML = '&uml;'  #: ¨
    COPY = '&copy;'  #: ©
    COPYRIGHT = '&copy;'  #: ©
    ORDF = '&ordf;'  #: ª
    LAQUO = '&laquo;'  #: «
    NOT = '&not;'  #: ¬
    # ­   &shy;   &#173;  Soft hyphen
    REG = '&reg;'  #: ®
    MACR = '&macr;'  #: ¯
    DEG = '&deg;'  #: °
    PLUSMN = '&plusmn;'  #: ±
    SUP2 = '&sup2;'  #: ²
    SUP3 = '&sup3;'  #: ³
    ACUTE = '&acute;'  #: ´
    MICRO = '&micro;'  #: µ
    PARA = '&para;'  #: ¶
    CEDIL = '&cedil;'  #: ¸
    SUP1 = '&sup1;'  #: ¹
    ORDM = '&ordm;'  #: º
    RAQUO = '&raquo;'  #: »
    FRAC14 = '&frac14;'  #: ¼
    FRAC12 = '&frac12;'  #: ½
    FRAC34 = '&frac34;'  #: ¾
    IQUEST = '&iquest;'  #: ¿
    TIMES = '&times;'  #: ×
    DIVIDE = '&divide;'  #: ÷

    # Math Symbols
    FORALL = '&forall;'  #: ∀
    PART = '&part;'  #: ∂
    EXIST = '&exist;'  #: ∃
    EMPTY = '&empty;'  #: ∅
    NABLA = '&nabla;'  #: ∇
    ISIN = '&isin;'  #: ∈
    NOTIN = '&notin;'  #: ∉
    NI = '&ni;'  #: ∋
    PROD = '&prod;'  #: ∏
    SUM = '&sum;'  #: ∑
    MINUS = '&minus;'  #: −
    LOWAST = '&lowast;'  #: ∗
    RADIC = '&radic;'  #: √
    PROP = '&prop;'  #: ∝
    INFIN = '&infin;'  #: ∞
    ANG = '&ang;'  #: ∠
    AND = '&and;'  #: ∧
    OR = '&or;'  #: ∨
    CAP = '&cap;'  #: ∩
    CUP = '&cup;'  #: ∪
    INT = '&int;'  #: ∫
    THERE4 = '&there4;'  #: ∴
    SIM = '&sim;'  #: ∼
    CONG = '&cong;'  #: ≅
    ASYMP = '&asymp;'  #: ≈
    NE = '&ne;'  #: ≠
    EQUIV = '&equiv;'  #: ≡
    LE = '&le;'  #: ≤
    GE = '&ge;'  #: ≥
    SUB = '&sub;'  #: ⊂
    SUP = '&sup;'  #: ⊃
    NSUB = '&nsub;'  #: ⊄
    SUBE = '&sube;'  #: ⊆
    SUPE = '&supe;'  #: ⊇
    OPLUS = '&oplus;'  #: ⊕
    OTIMES = '&otimes;'  #: ⊗
    PERP = '&perp;'  #: ⊥
    SDOT = '&sdot;'  #: ⋅

    # Greek Letters
    ALPHA = '&Alpha;'  #: Α
    BETA = '&Beta;'  #: Β
    GAMMA = '&Gamma;'  #: Γ
    DELTA = '&Delta;'  #: Δ
    EPSILON = '&Epsilon;'  #: Ε
    ZETA = '&Zeta;'  #: Ζ
    ETA = '&Eta;'  #: Η
    THETA = '&Theta;'  #: Θ
    IOTA = '&Iota;'  #: Ι
    KAPPA = '&Kappa;'  #: Κ
    LAMBDA = '&Lambda;'  #: Λ
    MU = '&Mu;'  #: Μ
    NU = '&Nu;'  #: Ν
    XI = '&Xi;'  #: Ξ
    OMICRON = '&Omicron;'  #: Ο
    PI = '&Pi;'  #: Π
    RHO = '&Rho;'  #: Ρ
    SIGMA = '&Sigma;'  #: Σ
    TAU = '&Tau;'  #: Τ
    UPSILON = '&Upsilon;'  #: Υ
    PHI = '&Phi;'  #: Φ
    CHI = '&Chi;'  #: Χ
    PSI = '&Psi;'  #: Ψ
    OMEGA = '&Omega;'  #: Ω
    ALPHA = '&alpha;'  #: α
    BETA = '&beta;'  #: β
    GAMMA = '&gamma;'  #: γ
    DELTA = '&delta;'  #: δ
    EPSILON = '&epsilon;'  #: ε
    ZETA = '&zeta;'  #: ζ
    ETA = '&eta;'  #: η
    THETA = '&theta;'  #: θ
    IOTA = '&iota;'  #: ι
    KAPPA = '&kappa;'  #: κ
    LAMBDA = '&lambda;'  #: λ
    MU = '&mu;'  #: μ
    NU = '&nu;'  #: ν
    XI = '&xi;'  #: ξ
    OMICRON = '&omicron;'  #: ο
    PI = '&pi;'  #: π
    RHO = '&rho;'  #: ρ
    SIGMAF = '&sigmaf;'  #: ς
    SIGMA = '&sigma;'  #: σ
    TAU = '&tau;'  #: τ
    UPSILON = '&upsilon;'  #: υ
    PHI = '&phi;'  #: φ
    CHI = '&chi;'  #: χ
    PSI = '&psi;'  #: ψ
    OMEGA = '&omega;'  #: ω
    THETASYM = '&thetasym;'  #: ϑ
    UPSIH = '&upsih;'  #: ϒ
    PIV = '&piv;'  #: ϖ

    OELIG = '&OElig;'  #: Œ
    oeLIG = '&oelig;'  #: œ
    SCARON = '&Scaron;'  #: Š
    Scaron = '&Scaron;'  #: Š
    scaron = '&scaron;'  #: š
    YUML = '&Yuml;'  #: Ÿ
    FNOF = '&fnof;'  #: ƒ
    CIRC = '&circ;'  #: ˆ
    TILDE = '&tilde;'  #: ˜

    #     &ensp;  &#8194; En space
    #     &emsp;  &#8195; Em space
    #     &thinsp;    &#8201; Thin space
    # ‌   &zwnj;  &#8204; Zero width non-joiner
    # ‍   &zwj;   &#8205; Zero width joiner
    # ‎   &lrm;   &#8206; Left-to-right mark
    # ‏   &rlm;   &#8207; Right-to-left mark

    NDASH = '&ndash;'  #: –
    MDASH = '&mdash;'  #: —
    LSQUO = '&lsquo;'  #: ‘
    RSQUO = '&rsquo;'  #: ’
    SBQUO = '&sbquo;'  #: ‚
    LDQUO = '&ldquo;'  #: “
    RDQUO = '&rdquo;'  #: ”
    BDQUO = '&bdquo;'  #: „
    DAGGER = '&dagger;'  #: †
    DAGGER = '&Dagger;'  #: ‡
    BULL = '&bull;'  #: •
    HELLIP = '&hellip;'  #: …
    PERMIL = '&permil;'  #: ‰
    PRIME = '&prime;'  #: ′
    PRIME = '&Prime;'  #: ″
    LSAQUO = '&lsaquo;'  #: ‹
    RSAQUO = '&rsaquo;'  #: ›
    OLINE = '&oline;'  #: ‾
    EURO = '&euro;'  #: €
    TRADE = '&trade;'  #: ™
    TRADEMARK = '&trade;'  #: ™

    # ARROWS
    LARR = '&larr;'  #: ←
    LEFT = '&larr;'  #: ←
    UARR = '&uarr;'  #: ↑
    UP = '&uarr;'  #: ↑
    RARR = '&rarr;'  #: →
    RIGHT = '&rarr;'  #: →
    DARR = '&darr;'  #: ↓
    DOWN = '&darr;'  #: ↓

    HARR = '&harr;'  #: ↔
    CRARR = '&crarr;'  #: ↵
    LCEIL = '&lceil;'  #: ⌈
    RCEIL = '&rceil;'  #: ⌉
    LFLOOR = '&lfloor;'  #: ⌊
    RFLOOR = '&rfloor;'  #: ⌋
    LOZ = '&loz;'  #: ◊

    SPADES = '&spades;'  #: ♠
    CLUBS = '&clubs;'  #: ♣
    HEARTS = '&hearts;'  #: ♥
    DIAMS = '&diams;'  #: ♦
    DIAMONDS = '&diams;'  #: ♦

    SUNG = '&sung;'  #: ♪
    FLAT = '&flat;'  #: ♭
    NATUR = '&natur;'  #: ♮
    NATURAL = '&natural;'  #: ♮
    SHARP = '&sharp;'  #: ♯

    CHECK = "&check;"  #: ✓
    CHECKMARK = "&checkmark;"  #: ✓
    TICK = "&check;"  #: ✓
    CROSS = "&cross;"  #: ✗

    OHM = '&ohm;'  #: Ω
    MHO = '&mho;'  #: ℧

    FRAC13 = '&frac13;'  #: ⅓
    FRAC23 = '&frac23;'  #: ⅔
    FRAC15 = '&frac15;'  #: ⅕
    FRAC25 = '&frac25;'  #: ⅖
    FRAC35 = '&frac35;'  #: ⅗
    FRAC45 = '&frac45;'  #: ⅘
    FRAC16 = '&frac16;'  #: ⅙
    FRAC56 = '&frac56;'  #: ⅚
    FRAC18 = '&frac18;'  #: ⅛
    FRAC38 = '&frac38;'  #: ⅜
    FRAC58 = '&frac58;'  #: ⅝
    FRAC78 = '&frac78;'  #: ⅞

    STAR = "&star;"  #: ☆
    STARF = "&starf;"  #: ★
    BIGSTAR = "&bigstar;"
    PHONE = "&phone;"  #: ☎
    FEMALE = "&female;"  #: ♀
    MALE = "&male;"  #: ♂



"""
    domonic.templates
    ====================================
    some builtin templates

"""

from domonic import *


class status_page():

    STATUS = {
        200: "200 OK",
        201: "201 Created",
        202: "202 Accepted",
        203: "203 Non-Authoritative Information",
        204: "204 No Content",
        205: "205 Reset Content",
        206: "206 Partial Content",
        207: "207 Multi-Status",
        208: "208 Already Reported",
        226: "226 IM Used",
        300: "300 Multiple Choices",
        301: "301 Moved Permanently",
        302: "302 Found",
        303: "303 See Other",
        304: "304 Not Modified",
        305: "305 Use Proxy",
        306: "306 Switch Proxy",
        307: "307 Temporary Redirect",
        308: "308 Permanent Redirect",
        400: "400 Bad Request",
        401: "401 Unauthorized",
        402: "402 Payment Required",
        403: "403 Forbidden",
        404: "404 Not Found",
        405: "405 Method Not Allowed",
        406: "406 Not Acceptable",
        407: "407 Proxy Authentication Required",
        408: "408 Request Timeout",
        409: "409 Conflict",
        410: "410 Gone",
        411: "411 Length Required",
        412: "412 Precondition Failed",
        413: "413 Payload Too Large",
        414: "414 URI Too Long",
        415: "415 Unsupported Media Type",
        416: "416 Range Not Satisfiable",
        417: "417 Expectation Failed",
        418: "418 I'm a teapot",
        421: "421 Misdirected Request",
        422: "422 Unprocessable Entity",
        423: "423 Locked",
        424: "424 Failed Dependency",
        426: "426 Upgrade Required",
        428: "428 Precondition Required",
        429: "429 Too Many Requests",
        431: "431 Request Header Fields Too Large",
        451: "451 Unavailable For Legal Reasons",
        499: "499 Client Closed Request",
        500: "500 Internal Server Error",
        501: "501 Not Implemented",
        502: "502 Bad Gateway",
        503: "503 Service Unavailable",
        504: "504 Gateway Timeout",
        505: "505 HTTP Version Not Supported",
        506: "506 Variant Also Negotiates",
        507: "507 Insufficient Storage",
        508: "508 Loop Detected",
        510: "510 Not Extended",
        511: "511 Network Authentication Required",

    }

    DEBUG_MODE = False

    def __init__(self, code=404, wholepage=True):
        self.status_code = code
        self.message = self.STATUS[code]
        # self.host = "http://localhost:8000"
        self.wholepage = wholepage

        self.status_node = div(_id="status")
        self.status_node += h1(f"{self.status_code}")
        self.status_node += p(self.message)

        # if status_page.DEBUG_MODE:
        #     import os
        #     self.status_node.appendChild(
        #         pre(os.environ)
        #     )
        #     import sys
        #     self.status_node.appendChild(
        #         pre(sys.path)
        #     )
        #     import traceback
        #     self.status_node.appendChild(
        #         pre(traceback.format_exc())
        #     )

        page = html(
            head(title(f"{self.status_code}")),
            body(self.status_node)
        )
        self.content = self.wholepage and page or self.status_node

    # def __getattr__(self, name):
    #     return getattr(self.status_node, name)

    # def __setattr__(self, name, value):
    #     setattr(self.status_node, name, value)

    def __str__(self):
        return f'{self.content}'

    # @staticmethod
    # def to_html(output_dir):
    #     for code, name in status_page.STATUS.items():
    #         page = status_page(code, name)
    #         with open(f"{output_dir}/{code}.html", "w") as f:
    #             f.write(f'{page}')


'''
class d_b():
    def __init__(self):
        pass
    def config(self, db_name, db_user, db_pass, db_host, db_port):
        self.db_name = db_name
        self.db_user = db_user
        self.db_pass = db_pass
        self.db_host = db_host
        self.db_port = db_port
    # def connect(self):
    # def query(self, query):
    # def close(self):
'''


# from domonic.templates import status_page

# @app.errorhandler(404)
# def page_not_found(e):
#     return status_page(404)



"""
    domonic.d3.polygon
    ====================================

"""

from domonic.javascript import Math, Array

def polygonArea(polygon):
    i = -1
    n = len(polygon)
    a = 0
    b = polygon[n - 1]
    area = 0

    while i < n:
        a = b
        b = polygon[i]
        area += a[1] * b[0] - a[0] * b[1]
        i+=1

    return area / 2


def polygonCentroid(polygon):
    i = -1
    n = len(polygon)
    x = 0
    y = 0
    a = 0
    b = polygon[n - 1]
    c = 0
    k = 0
    while i < n:
        a = b
        b = polygon[i]
        c = a[0] * b[1] - b[0] * a[1]
        k += c
        x += (a[0] + b[0]) * c
        y += (a[1] + b[1]) * c
        i+=1

    k *= 3
    return [x / k, y / k]


def cross(a, b, c):
    """[Returns the 2D cross product of AB and AC vectors, i.e., the z-component of
        the 3D cross product in a quadrant I Cartesian coordinate system (+x is
        right, +y is up). Returns a positive value if ABC is counter-clockwise,
        negative if clockwise, and zero if the points are collinear.]

    Args:
        a ([type]): [description]
        b ([type]): [description]
        c ([type]): [description]

    Returns:
        [type]: [description]
    """
    return (b[0] - a[0]) * (c[1] - a[1]) - (b[1] - a[1]) * (c[0] - a[0])


def lexicographicOrder(a, b, *args):
    return a[0] - b[0] or a[1] - b[1]


def computeUpperHullIndexes(points):
    """[Computes the upper convex hull per the monotone chain algorithm.
        Assumes points.length >= 3, is sorted by x, unique in y.
        Returns an array of indices into points in left-to-right order.]

    Args:
        points ([type]): [description]
    """
    n = len(points)
    indexes = [0, 1]
    size = 2
    for i in range(2, n):
        while (size > 1 and cross(points[indexes[size - 2]], points[indexes[size - 1]], points[i]) <= 0):
            size -= 1
        size += 1
        indexes[size] = i

    return Array(indexes).slice(0, size)  # remove popped points


def polygonHull(points):

    n = len(points)
    if n < 3:
        return None

    sortedPoints = []
    flippedPoints = []

    print(points)

    for i in range(0, n):
        print( points[i][0], points[i][1], i)
        sortedPoints.append([points[i][0], points[i][1], i])

    print(sortedPoints)

    sortedPoints = Array(sortedPoints).sort(lexicographicOrder)

    for i in range(0, n):
        flippedPoints[i] = [sortedPoints[i][0], -sortedPoints[i][1]]

    upperIndexes = computeUpperHullIndexes(sortedPoints)
    lowerIndexes = computeUpperHullIndexes(flippedPoints)

    # Construct the hull polygon, removing possible duplicate endpoints.
    skipLeft = lowerIndexes[0] == upperIndexes[0]
    skipRight = lowerIndexes[len(lowerIndexes) - 1] == upperIndexes[len(upperIndexes) - 1]
    hull = []

    # Add upper hull in right-to-l order.
    # Then add lower hull in left-to-right order.
    i = len(upperIndexes)
    while i >= 0:
        hull.append(points[sortedPoints[upperIndexes[i]][2]])
        i -= 1

    i = skipLeft
    while i < len(lowerIndexes) - skipRight:
        hull.append(points[sortedPoints[lowerIndexes[i]][2]])
        i += 1

    return hull


def polygonContains(polygon, point):
    n = len(polygon)
    p = polygon[n - 1]
    x = point[0]
    y = point[1]
    x0 = p[0]
    y0 = p[1]
    x1 = 0
    y1 = 0
    inside = False
    for i in range(0, n):
        p = polygon[i]
        x1 = p[0]
        y1 = p[1]
        if (((y1 > y) != (y0 > y)) and (x < (x0 - x1) * (y - y1) / (y0 - y1) + x1)):
            inside = not inside
        x0 = x1
        y0 = y1
    return inside


def polygonLength(polygon):
    i = -1
    n = len(polygon)
    b = polygon[n - 1]
    xa = 0
    ya = 0
    xb = b[0]
    yb = b[1]
    perimeter = 0
    while i < n:
        xa = xb
        ya = yb
        b = polygon[i]
        xb = b[0]
        yb = b[1]
        xa -= xb
        ya -= yb
        perimeter += Math.hypot(xa, ya)
        i += 1
    return perimeter



"""
    domonic.webapi.url
    ====================================
    https://developer.mozilla.org/en-US/docs/Web/API/URL

    # TODO - this will let us remove the javascript reference in the dom
    # TODO - move the unit tests for this class from javascript to webapi
    # TODO - untested

"""

import urllib


class URL(object):
    """ a-tag extends from URL """

    def __update__(self):
        # print( "update URL:", type(self), self  )
        try:
            # make obj with all old props
            new = {}
            new['protocol'] = self.url.scheme
            new['hostname'] = self.url.hostname
            new['href'] = self.url.geturl()
            new['port'] = self.url.port
            new['host'] = ''  # self.url.hostname
            new['pathname'] = self.url.path
            new['hash'] = ''  # self.url.hash
            new['search'] = ''  # self.url.hash

            # update it with all the new ones
            new = {}
            new['protocol'] = self.protocol
            new['hostname'] = self.hostname
            new['href'] = self.href
            new['port'] = self.port
            new['host'] = self.host
            new['pathname'] = self.pathname
            new['hash'] = self.hash  # self.hash
            new['search'] = self.search  # self.url.query
            new['_searchParams'] = self._searchParams  # URLSearchParams(self.url.query)
            # NOTE - rebuild happening here
            self.url = urllib.parse.urlsplit(
                new['protocol'] + "://" + new['host'] + new['pathname'] + new['hash'] + new['search'])

            self.href = self.url.geturl()

        except Exception:  # as e:
            # print('fails on props called by init as they dont exist yet')
            # print(e)
            pass

    def __init__(self, url: str = "", *args, **kwargs):  # TODO - relative to
        """URL

        builds a url

        Args:
            url (str): a url
        """
        self.url = urllib.parse.urlsplit(url)
        self.href = url  # self.url.geturl()
        self.protocol = self.url.scheme
        self.hostname = self.url.hostname
        self.port = self.url.port
        self.host = self.url.hostname
        self.pathname = self.url.path
        self.hash = ''
        self.search = self.url.query
        self._searchParams = URLSearchParams(self.url.query)

    @property
    def searchParams(self):
        return self._searchParams.toString()

    def toString(self):
        return str(self.href)

    # def toJson

    # @property
    # def href(self):
    # TODO - check js vs tag. does js version remove query?. if so detect self.
    #     return self.href

    # @href.setter
    # def href(self, href:str):
    #     self.url = href
    #     self.href = href

    @property
    def protocol(self):
        return self.__protocol

    @protocol.setter
    def protocol(self, p: str):
        self.__protocol = p
        # if self.ready : self.__update__() # TODO - this instead of silent err?
        self.__update__()

    @property
    def hostname(self):
        return self.__hostname

    @hostname.setter
    def hostname(self, h: str):
        if h is None:
            return
        if ":" in h:
            h = h.split(':')[0]
        self.__hostname = h
        self.__update__()

    @property
    def port(self):
        return self.__port

    @port.setter
    def port(self, p: str):
        self.__port = p
        self.__update__()

    @property
    def host(self):
        if self.port is not None:
            return self.hostname + ":" + str(self.port)
        else:
            return self.hostname

    @host.setter
    def host(self, h: str):
        if h is None:
            return
        p = self.port
        if ":" in h:
            p = int(h.split(':')[1])
            h = h.split(':')[0]
        self.__host = h
        self.hostname = h
        self.port = p
        self.__update__()

    @property
    def pathname(self):
        return self.__pathname

    @pathname.setter
    def pathname(self, p: str):
        self.__pathname = p
        self.__update__()

    @property
    def hash(self):
        """" hash Sets or returns the anchor part (#) of a URL """
        if '#' in self.href:
            return '#' + self.href.split('#')[1]
        # return ''
        return self.__hash

    @hash.setter
    def hash(self, h: str):
        self.__hash = h
        self.__update__()

    # @property
    # def origin(self):
        '''# origin    Returns the protocol, hostname and port number of a URL Location'''

    def __str__(self):
        return str(self.href)

    # NOTE - node -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- --
    # @staticmethod
    # def domainToASCII(domain: str):
    #     """[It returns the Punycode ASCII serialization of the domain.
    #     If domain is an invalid domain, the empty string is returned.]

    #     Args:
    #         domain (str): [description]
    #     """
    #     pass

    # @staticmethod
    # def domainToUnicode(domain: str):
    #     """[returns the Unicode serialization of the domain.
    #     If the domain is invalid, the empty string is returned]

    #     Args:
    #         domain (str): [description]
    #     """
    #     pass

    # @staticmethod
    # def fileURLToPath(url: str):
    #     """[ensures the correct decodings of percent-encoded characters as well as
    #     ensuring a cross-platform valid absolute path string.]

    #     Args:
    #         url (str): [The fully-resolved platform-specific file path.]
    #     """
    #     if url is None:
    #         return
    #     return urllib.parse.unquote(url)

    # @staticmethod
    # def format(URL, options):
    #     """[summary]

    #     Args:
    #         URL ([type]): [description]
    #         options ([type]): [description]
    #     """
    #     pass

    # @staticmethod
    # def pathToFileURL(path: str):
    #     """[summary]

    #     Args:
    #         path (str): [description]
    #     """
    #     pass

    # @staticmethod
    # def urlToHttpOptions(url: str):
    #     """[summary]

    #     Args:
    #         url (str): [description]
    #     """
    #     pass


class URLSearchParams:
    """[utility methods to work with the query string of a URL]
    """

    def __init__(self, paramString):  # , **paramsObj):
        """[Returns a URLSearchParams object instance.]

        Args:
            paramString ([type]): [ i.e. q=URLUtils.searchParams&topic=api]
        """
        # TODO - escape
        # import ast
        # TODO - dont think i can do this cant urls params have duplicate keys?
        # self.params = ast.literal_eval(paramString)
        if isinstance(paramString, str):
            if paramString.startswith('?'):
                paramString = paramString[1:len(paramString)]

            import urllib
            self.params = urllib.parse.parse_qs(paramString)
        elif hasattr(paramString, '__iter__'):
            self.params = [item for sublist in paramString for item in sublist]
        elif isinstance(paramString, dict):
            self.params = dict([(key, item) for key, item in paramString.iteritems()])
        else:
            raise TypeError("Malformed paramString.  Must be a string or a dict with dict like items. Got: %s" % paramString)

    def __iter__(self):
        for attr in self.params.items():  # dir(self.params.items()):
            # if not attr.startswith("__"):
            yield attr

    def append(self, key, value):
        """ Appends a specified key/value pair as a new search parameter """
        # TODO - ordereddict?
        self.params[key].append(value)  # [key]=value

    def delete(self, key):
        """ Deletes the given search parameter, and its associated value, from the list of all search parameters. """
        del self.params[key]

    def has(self, key):
        """ Returns a Boolean indicating if such a given parameter exists. """
        return key in self.params

    def entries(self):
        """ Returns an iterator allowing iteration through all key/value pairs contained in this object. """
        return self.params.items()

    def forEach(self, func):
        """ Allows iteration through all values contained in this object via a callback function. """
        for key, value in self.params.items():
            func(key, value)

    def keys(self):
        """ Returns an iterator allowing iteration through all keys of the key/value pairs contained in this object. """
        return self.params.keys()

    def get(self, key):
        """ Returns the first value associated with the given search parameter. """
        try:
            return self.params.get(key, None)[0]
        except Exception:
            return None

    def sort(self):
        """ Sorts all key/value pairs, if any, by their keys. """
        self.params.sort()

    def values(self):
        """ Returns an iterator allowing iteration through all values of the key/value pairs
        contained in this object. """
        return self.params.values()

    def toString(self):
        """ Returns a string containing a query string suitable for use in a URL. """
        # return '&'.join([str(x) for x in self.params])
        return urllib.parse.urlencode(self.params, doseq=True)
        # return str(self.params)

    def set(self, key, value):
        """ Sets the value associated with a given search parameter to the given value.
        If there are several values, the others are deleted. """
        self.params[key] = (value)

    def getAll(self, key):
        """ Returns all the values associated with a given search parameter. """
        return self.params.get(key)

    def __str__(self):
        return urllib.parse.urlencode(self.params, doseq=True)



"""
    test_webapi
    ~~~~~~~~~~~~~~~~
"""

import unittest

from domonic.dom import *
from domonic.javascript import *
from domonic.webapi import *
# from domonic.decorators import silence


class TestCase(unittest.TestCase):

    def test_encodingAPI(self):
        # utf8decoder = TextDecoder()  # default 'utf-8' or 'utf8'

        # u8arr = Uint8Array([240, 160, 174, 183])
        # i8arr = Int8Array([-16, -96, -82, -73])
        # u16arr = Uint16Array([41200, 47022])
        # i16arr = Int16Array([-24336, -18514])
        # i32arr = Int32Array([-1213292304])

        # console.log(utf8decoder.decode(u8arr))
        # console.log(utf8decoder.decode(i8arr))
        # console.log(utf8decoder.decode(u16arr))
        # console.log(utf8decoder.decode(i16arr))
        # console.log(utf8decoder.decode(i32arr))

        # win1251decoder = TextDecoder('windows-1251')
        # b = Uint8Array([207, 240, 232, 226, 229, 242, 44, 32, 236, 232, 240, 33])
        # console.log(win1251decoder.decode(b))  # // Привет, мир!
        pass

    def test_canvas(self):
        pass

    def test_clipboard(self):
        pass

    def test_cookiestore(self):
        pass

    def test_credentialmanagement(self):
        pass

    def test_crypto(self):
        pass

    def test_cssfontloading(self):
        pass

    def test_dragndrop(self):
        pass

    def test_filereader(self):
        pass

    def test_filesysmte(self):
        pass

    def test_fetch(self):
        pass

    def test_gamepad(self):
        pass

    def test_geolocation(self):
        pass

    def test_history(self):
        pass

    def test_gl(self):
        pass

    def test_intersectobserver(self):
        pass

    def test_mediacapabilities(self):
        pass

    def test_mediasession(self):
        pass

    def test_mediastream(self):
        pass

    def test_networkinfo(self):
        pass

    def test_notifications(self):
        pass

    def test_performance(self):
        pass

    def test_permissions(self):
        pass


if __name__ == '__main__':
    unittest.main()



"""
    domonic.d3.queue
    ====================================

    # TODO - completely untested

"""

# import {slice} from "./array";
from domonic.javascript import *


noabort = {}


class Queue():

    def __init__(self, size):
        self._size = size
        self._call = None
        self._error = None
        self._tasks = []
        self._data = []
        self._waiting = None
        self._active = None
        self._ended = None
        self._start = 0  # inside a synchronous task callback?

    def defer(self, callback, *args):
        if not type(callback, callable):
            raise Error("invalid callback")
        if (self._call):
            raise Error("defer after await")
        if (self._error != None):
            return self

        t = slice.call(args, 1)
        t.append(callback)
        self._waiting += 1
        self._tasks.append(t)
        poke(self)
        return self

    def abort(self):
        if self._error == None:
            abort(self, Error("abort"))
        return self

    def await(self, callback):
        if not type(callback, callable):
            raise Error("invalid callback")
        if (self._call):
            raise Error("multiple await")

        def _call(error, results):
            callback.apply(None, [error].concat(results))

        self._call = _call

        maybeNotify(self)
        return self

    def awaitAll(self, callback):
        if not type(callback, callable):
            raise Error("invalid callback")
        if (self._call):
            raise Error("multiple await")
        self._call = callback
        maybeNotify(self)
        return self


def poke(q):
    if not q._start:
        try:
            start(q)  # let the current task complete
        except Exception as e:
            if (q._tasks[q._ended + q._active - 1]):
                abort(q, e)  # task errored synchronously
            elif not q._data:
                raise e  # await callback errored synchronously


def start(q):
    q._start = q._waiting
    while (q._start and q._active < q._size):
        i = q._ended + q._active
        t = q._tasks[i]
        j = len(t) - 1
        c = t[j]
        t[j] = end(q, i)
        q._waiting -= 1
        q._active += 1
        t = c.apply(None, t)
        if not q._tasks[i]:
            continue  # task finished synchronously
        q._tasks[i] = t or noabort


def end(q, i):

    def _end(e, r):
        nonlocal q
        nonlocal i
        if not q._tasks[i]:
            return  # ignore multiple callbacks
        q._active -= 1
        q._ended += 1
        q._tasks[i] = None
        if q._error != None:
            return  # ignore secondary errors
        if e != None:
            abort(q, e)
        else:
            q._data[i] = r
            if q._waiting:
                poke(q)
            else:
                maybeNotify(q)

    return _end


def abort(q, e):
    i = len(q._tasks)
    q._error = e  # ignore active callbacks
    q._data = None  # allow gc
    q._waiting = None  # prevent starting

    while i >= 0:
        t = q._tasks[i]
        if t:
            q._tasks[i] = None
        if t.abort:
            try:
                t.abort()
            except Exception as e:
                pass
        i -= 1
    q._active = None  # allow notification
    maybeNotify(q)


def maybeNotify(q):
    if not q._active and q._call:
        d = q._data
        q._data = None  # allow gc
        q._call(q._error, d)


def queue(concurrency):
    if concurrency is None:
        concurrency = Infinity
    elif not (concurrency >= 1):
        raise Error("invalid concurrency")
    return Queue(concurrency)


"""
    domonic.ext.html5lib_
    ====================================

    stolen from here and modded to work with domonic instead of mindidom
    https://github.com/html5lib/html5lib-python/blob/master/html5lib/treebuilders/__init__.py

"""

'''
from __future__ import absolute_import, division, unicode_literals
# from .._utils import default_etree

treeBuilderCache = {}

def getTreeBuilder(treeType, implementation='domonic', **kwargs):
    treeType = treeType.lower()
    if treeType not in treeBuilderCache:
        if treeType == "domonic":
            from . import dom
            if implementation is None:
                from xml.dom import minidom
                implementation = minidom
            return dom.getDomModule(implementation, **kwargs).TreeBuilder
        else:
            raise ValueError("""domonic treebuilder required "%s" """ % treeType)
    return treeBuilderCache.get(treeType)
'''

# from __future__ import absolute_import, division, unicode_literals

try:
    from collections.abc import MutableMapping
except ImportError:  # Python 2.7
    from collections import MutableMapping
# from xml.dom import minidom, Node
import weakref

# from . import base
# from .. import constants
# from ..constants import namespaces
# from .._utils import moduleFactoryFactory

from html5lib.treebuilders import base
from html5lib import constants
from html5lib.constants import namespaces
from html5lib._utils import moduleFactoryFactory

from domonic.dom import DOMImplementation, Node


# def getDomBuilder(DomImplementation):
#     Dom = DomImplementation

def getDomBuilder(ignore):
    # Dom = DomImplementation

    class AttrList(MutableMapping):
        def __init__(self, element):
            self.element = element

        def __iter__(self):
            return iter(self.element.attributes.keys())

        def __setitem__(self, name, value):
            if isinstance(name, tuple):
                raise NotImplementedError
            else:
                attr = self.element.ownerDocument.createAttribute(name)
                attr.value = value
                self.element.attributes[name] = attr

        def __len__(self):
            return len(self.element.attributes)

        def items(self):
            # print(self.element, self.element.attributes)
            # return list(self.element.attributes.items())
            return list(self.element.attributes.items())
            # return self.element.attributes

        def values(self):
            return list(self.element.attributes.values())

        def __getitem__(self, name):
            if isinstance(name, tuple):
                raise NotImplementedError
            else:
                # print(self.element)
                # print(self.element.attributes, name)
                try:
                    return self.element.attributes[name].value
                except Exception as e:
                    print(e)
                    print('failed on :', name)
                    return ""

        def __delitem__(self, name):
            if isinstance(name, tuple):
                raise NotImplementedError
            else:
                del self.element.attributes[name]

    class NodeBuilder(base.Node):
        def __init__(self, element):
            base.Node.__init__(self, element.nodeName)
            self.element = element

        namespace = property(lambda self: hasattr(self.element, "namespaceURI") and
                             self.element.namespaceURI or None)

        def appendChild(self, node):
            node.parent = self
            self.element.appendChild(node.element)

        def insertText(self, data, insertBefore=None):
            text = self.element.ownerDocument.createTextNode(data)
            if insertBefore:
                self.element.insertBefore(text, insertBefore.element)
            else:
                self.element.appendChild(text)

        def insertBefore(self, node, refNode):
            self.element.insertBefore(node.element, refNode.element)
            node.parent = self

        def removeChild(self, node):
            if node.element.parentNode == self.element:
                self.element.removeChild(node.element)
            node.parent = None

        def reparentChildren(self, newParent):
            while self.element.hasChildNodes():
                child = self.element.firstChild
                self.element.removeChild(child)
                newParent.element.appendChild(child)
            self.childNodes = []

        def getAttributes(self):
            return AttrList(self.element)

        def setAttributes(self, attributes):
            if attributes:
                for name, value in list(attributes.items()):
                    if isinstance(name, tuple):
                        if name[0] is not None:
                            qualifiedName = (name[0] + ":" + name[1])
                        else:
                            qualifiedName = name[1]
                        self.element.setAttributeNS(name[2], qualifiedName,
                                                    value)
                    else:
                        self.element.setAttribute(
                            name, value)
        attributes = property(getAttributes, setAttributes)

        def cloneNode(self):
            return NodeBuilder(self.element.cloneNode(False))

        def hasContent(self):
            return self.element.hasChildNodes()

        def getNameTuple(self):
            if self.namespace is None:
                return namespaces["html"], self.name
            else:
                return self.namespace, self.name

        nameTuple = property(getNameTuple)

    class TreeBuilder(base.TreeBuilder):  # pylint:disable=unused-variable
        def documentClass(self):
            # self.dom = Dom.getDOMImplementation().createDocument(None, None, None)
            self.dom = DOMImplementation().createDocument(None, None, None)
            return weakref.proxy(self)

        def insertDoctype(self, token):
            name = token["name"]
            publicId = token["publicId"]
            systemId = token["systemId"]

            # domimpl = Dom.getDOMImplementation()
            domimpl = DOMImplementation()
            doctype = domimpl.createDocumentType(name, publicId, systemId)
            self.document.appendChild(NodeBuilder(doctype))
            # if Dom == minidom:
            doctype.ownerDocument = self.dom

        def elementClass(self, name, namespace=None):
            if namespace is None and self.defaultNamespace is None:
                node = self.dom.createElement(name)
            else:
                node = self.dom.createElementNS(namespace, name)

            return NodeBuilder(node)

        def commentClass(self, data):
            return NodeBuilder(self.dom.createComment(data))

        def fragmentClass(self):
            return NodeBuilder(self.dom.createDocumentFragment())

        def appendChild(self, node):
            self.dom.appendChild(node.element)

        def testSerializer(self, element):
            return testSerializer(element)

        def getDocument(self):
            return self.dom

        def getFragment(self):
            return base.TreeBuilder.getFragment(self).element

        def insertText(self, data, parent=None):
            data = data
            if parent != self:
                base.TreeBuilder.insertText(self, data, parent)
            else:
                # HACK: allow text nodes as children of the document node
                if hasattr(self.dom, '_child_node_types'):
                    # pylint:disable=protected-access
                    if Node.TEXT_NODE not in self.dom._child_node_types:
                        self.dom._child_node_types = list(self.dom._child_node_types)
                        self.dom._child_node_types.append(Node.TEXT_NODE)
                self.dom.appendChild(self.dom.createTextNode(data))

        from xml.dom import minidom
        implementation = minidom#DomImplementation
        name = None

    def testSerializer(element):
        element.normalize()
        rv = []

        def serializeElement(element, indent=0):
            if element.nodeType == Node.DOCUMENT_TYPE_NODE:
                if element.name:
                    if element.publicId or element.systemId:
                        publicId = element.publicId or ""
                        systemId = element.systemId or ""
                        rv.append("""|%s<!DOCTYPE %s "%s" "%s">""" %
                                  (' ' * indent, element.name, publicId, systemId))
                    else:
                        rv.append("|%s<!DOCTYPE %s>" % (' ' * indent, element.name))
                else:
                    rv.append("|%s<!DOCTYPE >" % (' ' * indent,))
            elif element.nodeType == Node.DOCUMENT_NODE:
                rv.append("#document")
            elif element.nodeType == Node.DOCUMENT_FRAGMENT_NODE:
                rv.append("#document-fragment")
            elif element.nodeType == Node.COMMENT_NODE:
                rv.append("|%s<!-- %s -->" % (' ' * indent, element.nodeValue))
            elif element.nodeType == Node.TEXT_NODE:
                rv.append("|%s\"%s\"" % (' ' * indent, element.nodeValue))
            else:
                if (hasattr(element, "namespaceURI") and
                        element.namespaceURI is not None):
                    name = "%s %s" % (constants.prefixes[element.namespaceURI],
                                      element.nodeName)
                else:
                    name = element.nodeName
                rv.append("|%s<%s>" % (' ' * indent, name))
                if element.hasAttributes():
                    attributes = []
                    for i in range(len(element.attributes)):
                        attr = element.attributes.item(i)
                        name = attr.nodeName
                        value = attr.value
                        ns = attr.namespaceURI
                        if ns:
                            name = "%s %s" % (constants.prefixes[ns], attr.localName)
                        else:
                            name = attr.nodeName
                        attributes.append((name, value))

                    for name, value in sorted(attributes):
                        rv.append('|%s%s="%s"' % (' ' * (indent + 2), name, value))
            indent += 2
            for child in element.childNodes:
                serializeElement(child, indent)
        serializeElement(element, 0)

        return "\n".join(rv)

    return locals()


# The actual means to get a module!
getDomModule = moduleFactoryFactory(getDomBuilder)


# class stupid_shit:
#     def getDOMImplementation(self):
#         from domonic.dom import DOMImplementation
        # return DOMImplementation


# if implementation is None:
from xml.dom import minidom
implementation = minidom

def getTreeBuilder():
    return getDomModule(implementation).TreeBuilder


"""
    domonic.d3
    ===================================

"""

from domonic.javascript import Array, String, RegExp, Object, Function

noop = {"value": lambda *args: {}}


def dispatch(*args):
    _ = {}
    for arg in args:
        if ((not isinstance(arg, str)) or (arg in _) or RegExp(r'[s.]').test(arg)):
            raise Exception("illegal type: " + arg)
        _[arg] = []
    return Dispatch(_)


def parseTypenames(typenames, types):

    def anon(t):
        name = ""
        i = String(t).indexOf(".")
        if i >= 0:
            name = String(t).slice(i + 1)
            t = String(t).slice(0, i)
        if types.get(t, None) == None:
            raise Exception("unknown type: " + t)
        return {"type": t, "name": name}

    import re
    return Array(re.split(r'/^|\s+/', String(typenames).trim())).map(anon)


class Dispatch():

    def __init__(self, _):
        self._ = _

    def on(self, typename, callback=None, *args):
        _ = self._
        T = parseTypenames(str(typename), _)
        t = None
        i = -1
        n = len(T)

        # If no callback was specified, return the callback of the given type and name.
        if callback is None:
            while i < n:
                typename = T[i]
                t = self.get(_[typename['type']], typename['name'])
                if t is not None:
                    return t
                i += 1
            return self

        # If a type was specified, set the callback for the given type and name.
        # Otherwise, if a None callback was specified, remove callbacks of the given name.
        if callback != None and not callable(callback):
            raise Exception("invalid callback: " + callback)
        while i < n:
            typename = T[i]
            if typename['type'] is not None:
                _[typename['type']] = self.set(_[typename['type']], typename['name'], callback)
            elif callback == None:
                for t in _:
                    _[typename['type']] = self.set(_[typename['type']], typename['name'], None)
            i += 1

        return self

    def copy(self):
        copy = {}
        _ = self._
        for t in _:
            copy[t] = Array(_[t]).slice()
        return Dispatch(copy)

    def call(self, type, that=None, *args):
        arguments = Array()
        n = len(args)
        if (n - 2) > 0:
            arguments = Array(n)
            for i in range(n):
                arguments[i] = arguments[i + 2]

        if not Object(self._).hasOwnProperty(type):
            raise Exception("unknown type: " + type)

        t = self._[type]
        for i in t:
            Function(i['value']).apply(that, arguments)

    def apply(self, type, that, *args):
        if not self._.hasOwnProperty(type):
            raise Exception("unknown type: " + type)
        t = self._[type]
        for i in t:
            Function(i['value']).apply(that, args)

    def get(self, type, name):
        n = len(type)
        for i in range(0, n):
            c = type[i]
            if c['name'] == name:
                return c['value']

    def set(self, type, name, callback):

        for i, t in enumerate(type):
            if t['name'] == name:
                t = noop
                type = Array(Array(type).slice(0, i)).concat(Array(type).slice(i + 1))
                if type is None:
                    type = []
                break

        if callback != None:
            type.append({"name": name, "value": callback})

        return type



"""
    test_cmd
    ~~~~~~~~~~~~~~~~
"""

import unittest

from domonic.cmd import dir
# from domonic.decorators import silence


class TestCase(unittest.TestCase):

    def test_cmd_dir(self):
        files = dir()
        print(files)
        # return
        self.assertTrue('domonic' in str(dir()))

        try:
            print(dir("..\\"))  # notice requires 2 backslashes
        except Exception:
            # will fail on non windows
            pass

        for line in dir():
            print("line:", line)


if __name__ == '__main__':
    unittest.main()



"""
    domonic.geom.shape.circle
    ====================================
    written by.ai
"""

from domonic.javascript import Math
from domonic.geom.shape import Shape


class Circle(Shape):

    def __init__(self, x, y, radius=1.0, color=None):
        super().__init__(color)
        self.radius = radius  # Create an instance variable radius

    @property
    def area(self):
        return self.radius * self.radius * Math.PI

    @property
    def perimeter(self):
        return 2 * self.radius * Math.PI

    @property
    def average_circumference(self):
        return 2 * self.radius

    @property
    def center(self):
        x, y = self.center = [self.radius, self.radius]
        return x, y

    @center.setter
    def center(self, center):
        self._center = center

    def __str__(self):
        return "Circle(%s, %s, %s)" % (self.center, self.radius, self.color)

    # def __repr__(self):
    #     return "Circle(%s, %s, %s)" % (self.center, self.radius, self.color)

    def __getstate__(self):
        return {'center': self.center, 'radius': self.radius}

    def __setstate__(self, state):
        self.center = state['center']
        self.radius = state['radius']

    def __copy__(self):
        return Circle(self.center, self.radius, self.color)

    def __deepcopy__(self, memo):
        return Circle(self.center, self.radius, self.color)

    def __contains__(self, other):
        return other in self.center

    def __len__(self):
        return len(self.center)

    def __getitem__(self, key):
        return self.center[key]

    def __setitem__(self, key, value):
        self.center[key] = value

    def __iter__(self):
        return iter(self.center)

    def __add__(self, other):
        return Circle(self.center + other.center, self.radius + other.radius)

    def __sub__(self, other):
        return Circle(self.center - other.center, self.radius - other.radius)

    def __mul__(self, other):
        return Circle(self.center * other, self.radius * other)

    # def __rmul__(self, other):
    #     return Circle(self.center * other, self.radius * other


class Oval(Circle):
    def __init__(self, radius=2.5, size=3):
        super().__init__(size, 'green')
        self.radius = radius
        self.x = self.width / 2 - self.radius / 2
        self.y = self.width / 2 + self.radius / 2


class Ellipse(Shape):
    def __init__(self, x, y, width, height, color=None):
        super().__init__(color)
        self.x = x
        self.y = y
        self.width = width
        self.height = height



"""
    test_domonic
    ~~~~~~~~~~~~
    - unit tests for domonic
    # TODO - tests for all bs5 pages
"""

import unittest
# import requests
# from mock import patch

from domonic import domonic

from domonic.html import *
from domonic.decorators import silence


class TestCase(unittest.TestCase):

    def test_hello_world(self):
        assert str(html(body(h1('Hello World!')))) == \
            '''<html><body><h1>Hello World!</h1></body></html>'''

    def test_html_attributes(self):
        assert str(div(_id='mydiv', _class='test', **{"_aria-label": True}, **{"_data-name": True}, _onclick="alert('hi');")) == \
            '''<div id="mydiv" class="test" aria-label="true" data-name="true" onclick="alert('hi');"></div>'''

        myel = div(_id='mydiv', _class='test', **{"_aria-label": True}, **{"_data-name": True}, _onclick="alert('hi');")
        assert myel.id == "mydiv"
        assert myel._id == "mydiv"
        assert myel._class == "test"
        assert myel.onclick == "alert('hi');"

    def test_create_element(self):
        # print(create_element('custom_el', div('some content'), _id="test"))
        assert str(create_element('custom_el', div('some content'), _id="test")) == \
            '''<custom_el id="test"><div>some content</div></custom_el>'''

    # @silence
    def test_domonic_parse(self):
        page = domonic.parse("<html><body>'some content'</body></html>")  # TODO - single comma
        page = domonic.parse("<html><body></body></html>")
        print(page)

    @silence
    def test_domonic_get(self):
        print("test_domonic_get-----------=-----------=-----------=-----------=-----------=-----------=-----------=")
        page = domonic.get("http://eventual.technology")
        page = domonic.get("https://v5.getbootstrap.com/docs/5.0/examples/checkout/")
        page = domonic.get("https://v5.getbootstrap.com/docs/5.0/examples/carousel/?#")
        page = domonic.get("https://v5.getbootstrap.com/docs/5.0/examples/dashboard/#")
        page = domonic.get("https://www.google.com")
        page = domonic.get("https://www.facebook.com")
        print("<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<")
        print(render(page))
        pass

    @silence
    def test_domonic_render(self):
        test = html(
            head(
            meta(_charset="utf-8"),
            # meta(**{"http-equiv": "X-UA-Compatible"},, _content="IE=edge"),
            title("eventual.technology"),
            meta(_name="viewport", _content="width=device-width, initial-scale=1"),
            meta(_name="description", _content="eventual technology"),
            meta(_name="keywords", _content="website, html5, javascript, python, software, aws"),
            meta(_name="author", _content="eventual.technology"),
            meta(_property="og:title", _content="eventual technology"),
            meta(_property="og:image", _content=""),
            meta(_property="og:url", _content=""),
            meta(_property="og:site_name", _content="eventual.technology"),
            meta(_property="og:description", _content=""),
            meta(_name="twitter:title", _content="eventual technology"),
            meta(_name="twitter:image", _content=""),
            meta(_name="twitter:url", _content="eventual.technology"),
            meta(_name="twitter:card", _content="")
            ),
            body(
                header(
                    h1(a(_href="mailto:mike@eventual.technology")),
                    h2("07535784121")
                ),
                footer(
                    img(_class="logo", _src="static/img/logo.svg", _alt="eventual technology")
                )

            )
        )
        # print(render(test))
        pass

    @silence
    def test_domonic_render_head(self):
        test = head(
            meta(_charset="utf-8"),
            # meta(**{"http-equiv": "X-UA-Compatible"},, _content="IE=edge"),
            title("byteface"),
            meta(_name="viewport", _content="width=device-width, initial-scale=1"),
            meta(_name="description", _content="eventual technology"),
            meta(_name="keywords", _content="website, html5, javascript, python, software, aws"),
            meta(_name="author", _content=""),

            meta(_property="og:title", _content=""),
            meta(_property="og:image", _content=""),
            meta(_property="og:url", _content=""),
            meta(_property="og:site_name", _content=""),
            meta(_property="og:description", _content=""),
            meta(_property="og:type", _content=""),

            meta(_name="twitter:title", _content=""),
            meta(_name="twitter:imag:src", _content=""),
            meta(_name="twitter:url", _content=""),
            meta(_name="twitter:card", _content=""),
            meta(_name="twitter:site", _content=""),
            meta(_name="twitter:description", _content=""),

            # link(_rel="dns-prefetch", _href="https://github.githubassets.com")
            # TOOD - list of 'classless'
            # link(_crossorigin="anonymous", _media="all", _integrity="sha512-1234", _rel="stylesheet", _href="styles.css")
            # link(_rel="search", _type="application/opensearchdescription+xml" href="/opensearch.xml" title="GitHub")

            meta(_property="fb:app_id", _content=""),
            meta(_name="google-site-verification", _content=""),
            meta(_name="google-analytics", _content=""),
            meta(_name="hostname", _content="site.com"),
            link(_rel="icon", _class="js-site-favicon", _type="image/png", _href="favicon.svg"),
            meta(_name="theme-color", _content="#1e2327"),
            link(_rel="manifest", _href="/manifest.json", _crossOrigin="use-credentials"),
            meta(_property="profile:username", _content="byteface")
        )
        print(render(test))
        pass

    @silence
    def test_domonic_render_bs5_checkout(self):
        test = html(
            head().html(
                meta(_charset="utf-8"),
                meta(_name="viewport", _content="width=device-width, initial-scale=1"),
                meta(_name="description", _content=""),
                meta(_name="author", _content="Mark Otto, Jacob Thornton, and Bootstrap contributors"),
                meta(_name="generator", _content="Hugo 0.72.0"),
                title("Checkout example · Bootstrap"),
                link(_rel="canonical", _href="https://v5.getbootstrap.com/docs/5.0/examples/checkout/"),
                link(_href="/docs/5.0/dist/css/bootstrap.min.css", _rel="stylesheet", __integrity="sha384-12345", __crossorigin="anonymous"),
                link(_rel="apple-touch-icon", _href="/docs/5.0/assets/img/favicons/apple-touch-icon.png", _sizes="180x180"),
                link(_rel="icon", _href="/docs/5.0/assets/img/favicons/favicon-32x32.png", _sizes="32x32", _type="image/png"),
                link(_rel="icon", _href="/docs/5.0/assets/img/favicons/favicon-16x16.png", _sizes="16x16", _type="image/png"),
                link(_rel="manifest", _href="/docs/5.0/assets/img/favicons/manifest.json"),
                link(_rel="mask-icon", _href="/docs/5.0/assets/img/favicons/safari-pinned-tab.svg", _color="#7952b3"),
                link(_rel="icon", _href="/docs/5.0/assets/img/favicons/favicon.ico"),
                meta(_name="theme-color", _content="#7952b3"),
                style("""
                .bd-placeholder-img {
                    font-size: 1.125rem;
                    text-anchor: middle;
                    -webkit-user-select: none;
                    -moz-user-select: none;
                    -ms-user-select: none;
                    user-select: none;
                }
                @media (min-width: 768px) {
                .bd-placeholder-img-lg {
                    font-size: 3.5rem;
                }
                }
                """),
                link(_href="form-validation.css", _rel="stylesheet"),
            ),
            body(_class="bg-light").html(  #, _html= # TODO - make an attribute to do the same
                div(_class="container").html(
                    div(_class="py-5 text-center").html(
                        img(_class="d-block mx-auto mb-4",
                            _src="/docs/5.0/assets/brand/bootstrap-solid.svg",
                            _alt="", _width="72", _height="72"
                            ),
                        h2("Checkout form"),
                        p("Below is an example form built entirely with Bootstrap’s form controls. \
                            Each required form group has a validation state that can be triggered \
                            by attempting to submit the form without completing it.",
                            _class="lead")
                    ),
                    div(_class="row g-3").html(
                        div(_class="col-md-5 col-lg-4 order-md-last").html(
                            h4(_class="d-flex justify-content-between align-items-center mb-3").html(
                                span("Your cart", _class="text-muted"),
                                span("3", _class="badge bg-secondary rounded-pill"),
                            ),
                            ul(_class="list-group mb-3").html(
                                li(_class="list-group-item d-flex justify-content-between lh-sm").html(
                                    div(
                                        h6("Product name", _class="my-0"),
                                        small("Brief description", _class="text-muted"),
                                    ),
                                    span("$12", _class="text-muted")
                                ),
                                li(_class="list-group-item d-flex justify-content-between lh-sm").html(
                                    div(
                                        h6("Second product", _class="my-0"),
                                        small("Brief description", _class="text-muted"),
                                    ),
                                    span("$8", _class="text-muted")
                                ),
                                li(_class="list-group-item d-flex justify-content-between lh-sm").html(
                                    div(
                                        h6("Third item", _class="my-0"),
                                        small("Brief description", _class="text-muted"),
                                        ),
                                    span("$5", _class="text-muted")
                                ),
                                li(_class="list-group-item d-flex justify-content-between bg-light").html(
                                    div(_class="text-success").html(
                                        h6("Promo code", _class="my-0"),
                                        small("EXAMPLECODE"),
                                        ),
                                    span("−$5", _class="text-success")
                                ),
                                li(_class="list-group-item d-flex justify-content-between").html(
                                    span("Total (USD)"),
                                    strong("$20")
                                )
                            ),
                            form(_class="card p-2").html(
                                div(_class="input-group").html(
                                    input(_type="text", _class="form-control", _placeholder="Promo code"),
                                    button("Redeem", _type="submit", _class="btn btn-secondary")
                                )
                            )
                        ),
                        div(_class="col-md-7 col-lg-8").html(
                            h4("Billing address", _class="mb-3"),
                            form(_class="needs-validation", _novalidate=True).html(
                                div(_class="row g-3").html(
                                    div(_class="col-sm-6").html(
                                        label("First name", _for="firstName", _class="form-label"),
                                        input(_type="text", _class="form-control", _id="firstName", _placeholder="", _value="", _required=True),
                                        div("Valid first name is required.", _class="invalid-feedback")
                                    ),
                                    div(_class="col-sm-6").html(
                                        label("Last name", _for="lastName", _class="form-label"),
                                        input(_type="text", _class="form-control", _id="lastName", _placeholder="", _value="", _required=True),
                                        div("Valid last name is required.", _class="invalid-feedback")
                                    ),
                                    div(_class="col-12").html(
                                        label("Username", _for="username", _class="form-label"),
                                        div(_class="input-group").html(
                                            span("@", _class="input-group-text"),
                                            input(_type="text", _class="form-control", _id="username", _placeholder="Username", _required=True),
                                            div("Your username is required.", _class="invalid-feedback")
                                        )
                                    ),
                                    div(_class="col-12").html(
                                        label("Email Optional", _for="email", _class="form-label"),
                                        input(_type="email", _class="form-control", _id="email", _placeholder="you@example.com"),
                                        div("Please enter a valid email address for shipping updates.", _class="invalid-feedback")
                                    ),
                                    div(_class="col-12").html(
                                        label("Address", _for="address", _class="form-label"),
                                        input(_type="text", _class="form-control", _id="address", _placeholder="1234 Main St", _required=True),
                                        div("Please enter your shipping address.", _class="invalid-feedback")
                                    ),
                                    div(_class="col-12").html(
                                        label("Address 2 Optional", _for="address2", _class="form-label"),
                                        input(_type="text", _class="form-control", _id="address2", _placeholder="Apartment or suite")
                                    ),
                                    div(_class="col-md-5").html(
                                        label("Country", _for="country", _class="form-label"),
                                        select(_class="form-select", _id="country", _required=True).html(
                                            option("Choose...", _value=""),
                                            option("United States"),
                                        ),
                                        div("Please select a valid country.", _class="invalid-feedback")
                                    ),
                                    div(_class="col-md-4").html(
                                        label("State", _for="state", _class="form-label"),
                                        select(_class="form-select", _id="state", _required=True).html(
                                            option("Choose...", _value=""),
                                            option("California"),
                                        ),
                                        div("Please provide a valid state.", _class="invalid-feedback")
                                    ),
                                    div(_class="col-md-3").html(
                                        label("Zip", _for="zip", _class="form-label"),
                                        input(_type="text", _class="form-control", _id="zip", _placeholder="", _required=True),
                                        div("Zip code required.", _class="invalid-feedback")
                                    )
                                ),
                                hr(_class="my-4"),
                                div(_class="form-check").html(
                                    input(_type="checkbox", _class="form-check-input", _id="same-address"),
                                    label("Shipping address is the same as my billing address", _class="form-check-label", _for="same-address"),
                                ),
                                div(_class="form-check").html(
                                    input(_type="checkbox", _class="form-check-input", _id="save-info"),
                                    label("Save this information for next time", _class="form-check-label", _for="save-info"),
                                ),
                                hr(_class="my-4"),
                                h4("Payment", _class="mb-3"),
                                div(_class="my-3").html(
                                    div(_class="form-check").html(
                                        input(_id="credit", _name="paymentMethod", _type="radio", _class="form-check-input", _checked=True, _required=True),
                                        label("Credit card", _class="form-check-label", _for="credit"),
                                    ),
                                    div(_class="form-check").html(
                                        input(_id="debit", _name="paymentMethod", _type="radio", _class="form-check-input", _required=True),
                                        label("Debit card", _class="form-check-label", _for="debit"),
                                    ),
                                    div(_class="form-check").html(
                                        input(_id="paypal", _name="paymentMethod", _type="radio", _class="form-check-input", _required=True),
                                        label("PayPal", _class="form-check-label", _for="paypal"),
                                    )
                                ),
                                div(_class="row gy-3").html(
                                    div(_class="col-md-6").html(
                                        label("Name on card", _for="cc-name", _class="form-label"),
                                        input(_type="text", _class="form-control", _id="cc-name", _placeholder="", _required=True),
                                        small("Full name as displayed on card", _class="text-muted"),
                                        div("Name on card is required", _class="invalid-feedback")
                                    ),
                                    div(_class="col-md-6").html(
                                        label("Credit card number", _for="cc-number", _class="form-label"),
                                        input(_type="text", _class="form-control", _id="cc-number", _placeholder="", _required=True),
                                        div("Credit card number is required", _class="invalid-feedback")
                                    ),
                                    div(_class="col-md-3").html(
                                        label("Expiration", _for="cc-expiration", _class="form-label"),
                                        input(_type="text", _class="form-control", _id="cc-expiration", _placeholder="", _required=True),
                                        div("Expiration date required", _class="invalid-feedback")
                                    ),
                                    div(_class="col-md-3").html(
                                        label("CVV", _for="cc-cvv", _class="form-label"),
                                        input(_type="text", _class="form-control", _id="cc-cvv", _placeholder="", _required=True),
                                        div("Security code required", _class="invalid-feedback")
                                    )
                                ),
                                hr(_class="my-4"),
                                button("Continue to checkout", _class="btn btn-primary btn-lg btn-block", _type="submit")
                            )
                        )
                    ),
                    footer(_class="my-5 pt-5 text-muted text-center text-small").html(
                        p("&copy; 2020 Company Name", _class="mb-1"),
                        ul(_class="list-inline").html(
                            li(a("Privacy", _href="#"), _class="list-inline-item"),
                            li(a("Terms", _href="#"), _class="list-inline-item"),
                            li(a("Support", _href="#"), _class="list-inline-item")
                        )
                    ),
                    script(_src="/docs/5.0/dist/js/bootstrap.bundle.min.js", _integrity="sha384-12345", _crossorigin="anonymous"),
                    script(_src="form-validation.js")
                )
            )
        )

        # print(render(test,'bs5_test_checkout.html'))
        print(render(test))
        pass

    @silence
    def test_domonic_render_bs5_carousel(self):
        test = html(_lang="en").html(
            head(
                meta(_charset="utf-8"),
                meta(_name="viewport", _content="_width=device-width, initial-scale=1"),
                meta(_name="description", _content=""),
                meta(_name="author", _content="Mark Otto, Jacob Thornton, and Bootstrap contributors"),
                meta(_name="generator", _content="Hugo 0.72.0"),
                title("Carousel Template · Bootstrap"),
                link(_rel="canonical", _href="https://v5.getbootstrap.com/docs/5.0/examples/carousel/"),
                link(_href="/docs/5.0/dist/css/bootstrap.min.css", _rel="stylesheet", _integrity="sha384-1234", _crossorigin="anonymous"),
                link(_rel="apple-touch-icon", _href="/docs/5.0/assets/img/favicons/apple-touch-icon.png", _sizes="180x180"),
                link(_rel="icon", _href="/docs/5.0/assets/img/favicons/favicon-32x32.png", _sizes="32x32", _type="image/png"),
                link(_rel="icon", _href="/docs/5.0/assets/img/favicons/favicon-16x16.png", _sizes="16x16", _type="image/png"),
                link(_rel="manifest", _href="/docs/5.0/assets/img/favicons/manifest.json"),
                link(_rel="mask-icon", _href="/docs/5.0/assets/img/favicons/safari-pinned-tab.svg", _color="#7952b3"),
                link(_rel="icon", _href="/docs/5.0/assets/img/favicons/favicon.ico"),
                meta(_name="theme-color", _content="#7952b3"),
                style("""
                .bd-placeholder-img {
                    font-size: 1.125rem;
                    text-anchor: middle;
                    -webkit-user-select: none;
                    -moz-user-select: none;
                    -ms-user-select: none;
                    user-select: none;
                }
                @media (min-width: 768px) {
                    .bd-placeholder-img-lg {
                    font-size: 3.5rem;
                    }
                }
                """),
                link(_href="carousel.css", _rel="stylesheet"),
            ),
            body(
                header(
                    nav(_class="navbar navbar-expand-md navbar-dark fixed-top bg-dark").html(
                        div(_class="container-fluid").html(
                            a("Carousel", _class="navbar-brand", _href="#"),
                            button(span(_class="navbar-toggler-icon"), _class="navbar-toggler", _type="button", **{"_data-toggle": "collapse"}, **{"_data-target": "#navbarCollapse"}, **{"_aria-controls": "navbarCollapse"}, **{"_aria-expanded": "false"}, **{"_aria-label": "Toggle navigation"})
                        ),
                        div(_class="collapse navbar-collapse", _id="navbarCollapse").html(
                                ul(_class="navbar-nav mr-auto mb-2 mb-md-0").html(
                                    li(a("Home", _class="nav-link", **{"_aria-current": "page"}, _href="#"), _class="nav-item active"),
                                    li(a("Link", _class="nav-link", _href="#"), _class="nav-item"),
                                    li(a("Disabled", _class="nav-link disabled", _href="#", _tabindex="-1", **{"_aria-disabled": "true"}),
                                    _class="nav-item"
                                    )
                                ),
                                form(_class="d-flex").html(
                                    input(button("Search", _class="btn btn-outline-success", _type="submit"),
                                    _class="form-control mr-2", _type="search", _placeholder="Search", **{"_aria-label": "Search"},
                                )
                            )
                        )
                    )
                ),
                main(
                    div(_id="myCarousel", _class="carousel slide", **{"_data-ride": "carousel"}).html(
                    ol(_class="carousel-indicators").html(
                        li(**{"_data-target": "#myCarousel"}, **{"_data-slide-to": "0"}, _class="active"),
                        li(**{"_data-target": "#myCarousel"}, **{"_data-slide-to": "1"}),
                        li(**{"_data-target": "#myCarousel"}, **{"_data-slide-to": "2"})
                    ),
                    div(_class="carousel-inner").html(
                    div(_class="carousel-item active").html(
                        """<svg _class="bd-placeholder-img", _width="100%", _height="100%" xmlns="http://www.w3.org/2000/svg" preserveAspectRatio="xMidYMid slice", _role="img" focusable="false",<rect _width="100%", _height="100%" fill="#777")</svg>""",
                        div(_class="container").html(
                            div(_class="carousel-caption text-left").html(
                                h1("Example headline."),
                                p("Cras justo odio, dapibus ac facilisis in, egestas eget quam. Donec id elit non mi porta gravida at eget metus. Nullam id dolor id nibh ultricies vehicula ut id elit."),
                                p(a("Sign up today", _class="btn btn-lg btn-primary", _href="#", _role="button"))
                            )
                        )
                    ),
                    div(_class="carousel-item").html(
                        """<svg _class="bd-placeholder-img", _width="100%", _height="100%" xmlns="http://www.w3.org/2000/svg" preserveAspectRatio="xMidYMid slice", _role="img" focusable="false",<rect _width="100%", _height="100%" fill="#777")</svg>""",
                        div(_class="container").html(
                            div(_class="carousel-caption").html(
                                h1("Another example headline."),
                                p("Cras justo odio, dapibus ac facilisis in, egestas eget quam. Donec id elit non mi porta gravida at eget metus. Nullam id dolor id nibh ultricies vehicula ut id elit."),
                                p(a("Learn more", _class="btn btn-lg btn-primary", _href="#", _role="button"))
                            )
                        )
                    ),
                    div(_class="carousel-item").html(
                        """<svg _class="bd-placeholder-img", _width="100%", _height="100%" xmlns="http://www.w3.org/2000/svg" preserveAspectRatio="xMidYMid slice", _role="img" focusable="false",<rect _width="100%", _height="100%" fill="#777")</svg>""",
                        div(_class="container").html(
                            div(_class="carousel-caption text-right").html(
                                h1("One more for good measure."),
                                p("Cras justo odio, dapibus ac facilisis in, egestas eget quam. Donec id elit non mi porta gravida at eget metus. Nullam id dolor id nibh ultricies vehicula ut id elit."),
                                p(a("Browse gallery", _class="btn btn-lg btn-primary", _href="#", _role="button"))
                            )
                        )
                    )
                    ),
                    a(_class="carousel-control-prev", _href="#myCarousel", _role="button", **{"_data-slide": "prev"}).html(
                        span(**{"_aria-hidden": "true"}, _class="carousel-control-prev-icon"),
                        span("Previous", _class="sr-only")
                    ),
                    a(_class="carousel-control-next", _href="#myCarousel", _role="button", **{"_data-slide": "next"}).html(
                        span(_class="carousel-control-next-icon", **{"_aria-hidden": "true"}),
                        span("Next", _class="sr-only")
                    )
                ),
                div(_class="container marketing").html(
                    div(_class="row").html(
                    div(_class="col-lg-4").html(
                        """<svg _class="bd-placeholder-img rounded-circle", _width="140", _height="140" xmlns="http://www.w3.org/2000/svg" aria-label="Placeholder: 140x140" preserveAspectRatio="xMidYMid slice", _role="img" focusable="false",title(Placeholder)<rect _width="100%", _height="100%" fill="#777")<text x="50%" y="50%" fill="#777" dy=".3em">140x140</text></svg>""",
                        h2("Heading"),
                        p("Donec sed odio dui. Etiam porta sem malesuada magna mollis euismod. Nullam id dolor id nibh ultricies vehicula ut id elit. Morbi leo risus, porta ac consectetur ac, vestibulum at eros. Praesent commodo cursus magna."),
                        p(a("View details &raquo;", _class="btn btn-secondary", _href="#", _role="button"))
                    ),
                    div(_class="col-lg-4").html(
                        """<svg _class="bd-placeholder-img rounded-circle", _width="140", _height="140" xmlns="http://www.w3.org/2000/svg" aria-label="Placeholder: 140x140" preserveAspectRatio="xMidYMid slice", _role="img" focusable="false",title(Placeholder)<rect _width="100%", _height="100%" fill="#777")<text x="50%" y="50%" fill="#777" dy=".3em">140x140</text></svg>""",
                        h2("Heading"),
                        p("Duis mollis, est non commodo luctus, nisi erat porttitor ligula, eget lacinia odio sem nec elit. Cras mattis consectetur purus sit amet fermentum. Fusce dapibus, tellus ac cursus commodo, tortor mauris condimentum nibh."),
                        p(a("View details &raquo;", _class="btn btn-secondary", _href="#", _role="button"))
                    ),
                    div(_class="col-lg-4").html(
                        """<svg _class="bd-placeholder-img rounded-circle", _width="140", _height="140" xmlns="http://www.w3.org/2000/svg" aria-label="Placeholder: 140x140" preserveAspectRatio="xMidYMid slice", _role="img" focusable="false",title(Placeholder)<rect _width="100%", _height="100%" fill="#777")<text x="50%" y="50%" fill="#777" dy=".3em">140x140</text></svg>""",
                        h2("Heading"),
                        p("Donec sed odio dui. Cras justo odio, dapibus ac facilisis in, egestas eget quam. Vestibulum id ligula porta felis euismod semper. Fusce dapibus, tellus ac cursus commodo, tortor mauris condimentum nibh, ut fermentum massa justo sit amet risus."),
                        p(a("View details &raquo;", _class="btn btn-secondary", _href="#", _role="button")
                    ),
                    hr(_class="featurette-divider"),
                    div(_class="row featurette").html(
                        div(_class="col-md-7").html(
                            h2("First featurette heading.", span(_class="text-muted"), "It’ll blow your mind.", _class="featurette-heading"),
                            p("Donec ullamcorper nulla non metus auctor fringilla. Vestibulum id ligula porta felis euismod semper. Praesent commodo cursus magna, vel scelerisque nisl consectetur. Fusce dapibus, tellus ac cursus commodo.",
                            _class="lead")
                        ),
                        div("""<svg _class="bd-placeholder-img bd-placeholder-img-lg featurette-image img-fluid mx-auto", _width="500", _height="500" xmlns="http://www.w3.org/2000/svg" aria-label="Placeholder: 500x500" preserveAspectRatio="xMidYMid slice", _role="img" focusable="false",title(Placeholder)<rect _width="100%", _height="100%" fill="#eee")<text x="50%" y="50%" fill="#aaa" dy=".3em">500x500</text></svg>""",
                            _class="col-md-5"
                        )
                    ),
                    hr(_class="featurette-divider"),
                    div(_class="row featurette").html(
                        div(_class="col-md-7 order-md-2").html(
                            h2("Oh yeah, it’s that good.", span(_class="text-muted"), "See for yourself.", _class="featurette-heading")),
                            p("Donec ullamcorper nulla non metus auctor fringilla. Vestibulum id ligula porta felis euismod semper. Praesent commodo cursus magna, vel scelerisque nisl consectetur. Fusce dapibus, tellus ac cursus commodo.", _class="lead")
                        ),
                        div(_class="col-md-5 order-md-1").html(
                            """<svg _class="bd-placeholder-img bd-placeholder-img-lg featurette-image img-fluid mx-auto", _width="500", _height="500" xmlns="http://www.w3.org/2000/svg" aria-label="Placeholder: 500x500" preserveAspectRatio="xMidYMid slice", _role="img" focusable="false",title(Placeholder)<rect _width="100%", _height="100%" fill="#eee")<text x="50%" y="50%" fill="#aaa" dy=".3em">500x500</text></svg>"""
                        )
                    ),
                    hr(_class="featurette-divider"),
                    div(_class="row featurette").html(
                    div(_class="col-md-7").html(
                        h2("Checkmate.", "And lastly, this one.", span(_class="text-muted"), _class="featurette-heading"),
                        p("Donec ullamcorper nulla non metus auctor fringilla. Vestibulum id ligula porta felis euismod semper. Praesent commodo cursus magna, vel scelerisque nisl consectetur. Fusce dapibus, tellus ac cursus commodo.",
                        _class="lead")
                    ),
                    div("""<svg _class="bd-placeholder-img bd-placeholder-img-lg featurette-image img-fluid mx-auto", _width="500", _height="500" xmlns="http://www.w3.org/2000/svg" aria-label="Placeholder: 500x500" preserveAspectRatio="xMidYMid slice", _role="img" focusable="false",title(Placeholder)<rect _width="100%", _height="100%" fill="#eee")<text x="50%" y="50%" fill="#aaa" dy=".3em">500x500</text></svg>""",
                        _class="col-md-5"
                        )
                    ),
                    hr(_class="featurette-divider"),
                    footer(_class="container").html(
                        p("Back to top", a(_href="#"), _class="float-right"),
                        p("&copy; 2017-2020 Company, Inc. &middot;", a("Privacy", _href="#"), "&middot;", a("Terms", _href="#"))
                    )
                    ),
                    script(_src="/docs/5.0/dist/js/bootstrap.bundle.min.js", _integrity="sha384-12345", _crossorigin="anonymous")
                    )
                )
            )
        )
        # print(render(test,'bs5_test_carousel.html'))
        print(render(test))
        pass


    @silence
    def test_domonic_render_bs5_dashboard(self):
        root = html(_lang="en")
        test = root.html(
            head(
                meta(_charset="utf-8"),
                meta(_name="viewport", _content="width=device-width, initial-scale=1"),
                meta(_name="description", _content=""),
                meta(_name="author", _content="Mark Otto, Jacob Thornton, and Bootstrap contributors"),
                meta(_name="generator", _content="Hugo 0.72.0"),
                title("Dashboard Template · Bootstrap"),
                link(_rel="canonical", _href="https://v5.getbootstrap.com/docs/5.0/examples/dashboard/"),
                link(_href="/docs/5.0/dist/css/bootstrap.min.css", _rel="stylesheet", _integrity="sha384-1234", _crossorigin="anonymous"),
                link(_rel="apple-touch-icon", _href="/docs/5.0/assets/img/favicons/apple-touch-icon.png", _sizes="180x180"),
                link(_rel="icon", _href="/docs/5.0/assets/img/favicons/favicon-32x32.png", _sizes="32x32", _type="image/png"),
                link(_rel="icon", _href="/docs/5.0/assets/img/favicons/favicon-16x16.png", _sizes="16x16", _type="image/png"),
                link(_rel="manifest", _href="/docs/5.0/assets/img/favicons/manifest.json"),
                link(_rel="mask-icon", _href="/docs/5.0/assets/img/favicons/safari-pinned-tab.svg", _color="#7952b3"),
                link(_rel="icon", _href="/docs/5.0/assets/img/favicons/favicon.ico"),
                meta(_name="theme-color", _content="#7952b3"),
                style("""
                .bd-placeholder-img {
                    font-size: 1.125rem;
                    text-anchor: middle;
                    -webkit-user-select: none;
                    -moz-user-select: none;
                    -ms-user-select: none;
                    user-select: none;
                }
                @media (min-width: 768px) {
                .bd-placeholder-img-lg {
                    font-size: 3.5rem;
                }
                }
                """),
                link(_href="dashboard.css", _rel="stylesheet"),
            ),
            body(
                nav(_class="navbar navbar-dark sticky-top bg-dark flex-md-nowrap p-0 shadow").html(
                    a("Company name", _class="navbar-brand col-md-3 col-lg-2 mr-0 px-3", _href="#"),
                    button(span(_class="navbar-toggler-icon"), 
                        _class="navbar-toggler position-absolute d-md-none collapsed",
                        _type="button", 
                        **{"_data-toggle": "collapse"}, 
                        **{"_data-target": "#sidebarMenu"}, 
                        **{"_aria-controls": "sidebarMenu"}, 
                        **{"_aria-expanded": "false"}, 
                        **{"_aria-label": "Toggle navigation"}
                    ),
                    input(_class="form-control form-control-dark w-100", _type="text", _placeholder="Search", **{"_aria-label": "Search"}).html(
                        ul(_class="navbar-nav px-3").html(
                            li(a("Sign out",_class="nav-link", _href="#"), _class="nav-item text-nowrap")
                        )
                    ),
                    div(_class="container-fluid").html(
                    div(_class="row").html(
                        nav(_id="sidebarMenu", _class="col-md-3 col-lg-2 d-md-block bg-light sidebar collapse").html(
                            div(_class="position-sticky pt-3").html(
                                ul(_class="nav flex-column").html(
                                    li(
                                        a(span(**{"_data-feather": "home"}), "Dashboard", _class="nav-link active", **{"_aria-current": "page"}, _href="#"),
                                        _class="nav-item"
                                    ),
                                    li(
                                        a(span(**{"_data-feather": "file"}), "Orders", _class="nav-link", _href="#"),
                                        _class="nav-item"
                                    ),
                                    li(
                                        a(span(**{"_data-feather": "shopping-cart"}), "Products", _class="nav-link", _href="#"),
                                        _class="nav-item"
                                    ),
                                    li(
                                        a(span(**{"_data-feather": "users"}), "Customers", _class="nav-link", _href="#"),
                                        _class="nav-item"
                                    ),
                                    li(
                                        a(span(**{"_data-feather": "bar-chart-2"}), "Reports", _class="nav-link", _href="#"),
                                        _class="nav-item"
                                    ),
                                    li(
                                        a(span(**{"_data-feather": "layers"}),"Integrations",_class="nav-link", _href="#"),
                                        _class="nav-item"
                                    )
                                )
                            ),
                            h6(_class="sidebar-heading d-flex justify-content-between align-items-center px-3 mt-4 mb-1 text-muted").html(
                            span("Saved reports"),
                                a(span(**{"_data-feather": "plus-circle"}, **{"_aria-label": "Add a new report"}),
                                    _class="link-secondary", _href="#")
                                )
                            ),
                            ul(_class="nav flex-column mb-2").html(
                                li(
                                    a(span(**{"_data-feather": "file-text"}), "Current month", _class="nav-link", _href="#"),
                                    _class="nav-item"
                                ),
                                li(
                                    a(span(**{"_data-feather": "file-text"}), "Last quarter", _class="nav-link", _href="#"),
                                    _class="nav-item"
                                ),
                                li(
                                    a(span(**{"_data-feather": "file-text"}), "Social engagement", _class="nav-link", _href="#"),
                                    _class="nav-item"
                                ),
                                li(
                                    a(span(**{"_data-feather": "file-text"}), "Year-end sale", _class="nav-link", _href="#"),
                                    _class="nav-item"
                                )
                            )
                        )
                    ),
                    main(_class="col-md-9 ml-sm-auto col-lg-10 px-md-4").html(
                        div(_class="d-flex justify-content-between flex-wrap flex-md-nowrap align-items-center pt-3 pb-2 mb-3 border-bottom").html(
                            h1("Dashboard", _class="h2"),
                            div(_class="btn-toolbar mb-2 mb-md-0").html(
                                div(_class="btn-group mr-2").html(
                                    button("Share", _type="button", _class="btn btn-sm btn-outline-secondary"),
                                    button("Export", _type="button", _class="btn btn-sm btn-outline-secondary")
                                ),
                                button(span(**{"_data-feather": "calendar"}), "This week", _type="button", _class="btn btn-sm btn-outline-secondary dropdown-toggle")
                            )
                        ),
                        canvas(_class="my-4 w-100", _id="myChart", _width="900", _height="380"),
                        h2("Section title"),
                        div(_class="table-responsive").html(
                            table(_class="table table-striped table-sm").html(
                                thead(
                                    tr(
                                        th("  #"),
                                        th("Header"),
                                        th("Header"),
                                        th("Header"),
                                        th("Header")
                                        )
                                    ),
                                tbody(
                                    tr(
                                        td("1, 001"),
                                        td("Lorem"),
                                        td("ipsum"),
                                        td("dolor"),
                                        td("sit")
                                    ),
                                    tr(
                                        td("1, 002"),
                                        td("amet"),
                                        td("consectetur"),
                                        td("adipiscing"),
                                        td("elit")
                                    ),
                                    tr(
                                        td("1, 003"),
                                        td("Integer"),
                                        td("nec"),
                                        td("odio"),
                                        td("Praesent")
                                    ),
                                    tr(
                                        td("1, 003"),
                                        td("libero"),
                                        td("Sed"),
                                        td("cursus"),
                                        td("ante")
                                    ),
                                    tr(
                                        td("1, 004"),
                                        td("dapibus"),
                                        td("diam"),
                                        td("Sed"),
                                        td("nisi")
                                    ),
                                    tr(
                                        td("1, 005"),
                                        td("Nulla"),
                                        td("quis"),
                                        td("sem"),
                                        td("at")
                                    ),
                                    tr(
                                        td("1, 006"),
                                        td("nibh"),
                                        td("elementum"),
                                        td("imperdiet"),
                                        td("Duis")
                                    ),
                                    tr(
                                        td("1, 007"),
                                        td("sagittis"),
                                        td("ipsum"),
                                        td("Praesent"),
                                        td("mauris")
                                    ),
                                    tr(
                                        td("1, 008"),
                                        td("Fusce"),
                                        td("nec"),
                                        td("tellus"),
                                        td("sed")
                                    ),
                                    tr(
                                        td("1, 009"),
                                        td("augue"),
                                        td("semper"),
                                        td("porta"),
                                        td("Mauris")
                                    ),
                                    tr(
                                        td("1, 010"),
                                        td("massa"),
                                        td("Vestibulum"),
                                        td("lacinia"),
                                        td("arcu")
                                    ),
                                    tr(
                                        td("1, 011"),
                                        td("eget"),
                                        td("nulla"),
                                        td("Class"),
                                        td("aptent")
                                    ),
                                    tr(
                                        td("1, 012"),
                                        td("taciti"),
                                        td("sociosqu"),
                                        td("ad"),
                                        td("litora")
                                    ),
                                    tr(
                                        td("1, 013"),
                                        td("torquent"),
                                        td("per"),
                                        td("conubia"),
                                        td("nostra")
                                    ),
                                    tr(
                                        td("1, 014"),
                                        td("per"),
                                        td("inceptos"),
                                        td("himenaeos"),
                                        td("Curabitur")
                                    ),
                                    tr(
                                        td("1, 015"),
                                        td("sodales"),
                                        td("ligula"),
                                        td(" in "),
                                        td("libero")
                                    )
                                )
                            )
                        )
                    )
                )
            ),
            script(_src="/docs/5.0/dist/js/bootstrap.bundle.min.js", _integrity="sha384-1234", _crossorigin="anonymous"),
            script(_src="https://cdnjs.cloudflare.com/ajax/libs/feather-icons/4.24.1/feather.min.js", _integrity="sha384-1234", _crossorigin="anonymous"),
            script(_src="https://cdnjs.cloudflare.com/ajax/libs/Chart.js/2.9.3/Chart.min.js", _integrity="sha384-1234", _crossorigin="anonymous"),
            script(_src="dashboard.js")
        )

        # print(render(test,'bs5_test_dashboard.html'))
        # print(render(test))

        # document.getElementsByTagName("div")[0].getAttributeNode("class");
        # print(test.getElementsByTagName("div") )

        # test = html().html(head(),body())
        # print(type(root))
        # print(root.getElementsByTagName("button")[0].getAttribute("class"))
        # print(render(root))
        pass


    def test_domonic_render_a_tag(self):
        atag = a(_href="https://somesite.com:8000/blog/article-one#some-hash")
        assert str(atag) == '<a href="https://somesite.com:8000/blog/article-one#some-hash"></a>'
        assert "https://somesite.com:8000/blog/article-one#some-hash" == atag.href
        assert 'https' == atag.protocol
        assert '8000' == str(atag.port)
        atag.protocol = "http"
        assert 'http' == atag.protocol
        assert "http://somesite.com:8000/blog/article-one#some-hash" == atag.href
        atag.port = 8983
        assert '8983' == str(atag.port)
        assert "http://somesite.com:8983/blog/article-one#some-hash" == atag.href
        assert 'somesite.com' == atag.hostname
        # print('host:',atag.host)


    def test_domonic_render_a_tag_query_params(self):
        # from domonic import a, render
        urls = [
        'example.com/stuff?things=morestuff',
        'https://example.com/stuff?things=morestuff',
        'https://example.com/stuff',
        'https://www.example.com/stuff?thing',
        'https://www.example.com/?stuff'
        ]
        for url in urls:
            print( 'zz',url)
            print( 'zz2',render(a(_href=url)) )
            assert f'''{render(a(_href=url))}''' == f'''<a href="{url}"></a>'''

        # a tag no href TODO
        # print( render(a(_name="test")) )


    def test_with(self):

        # d = 
        # print(d.head)
        # with html() as d:
        #     print(d)
        #     print(type(d))
        #     d += div('test')
        #     with div(_class="container"):
        #         p('test2')
        #         p('test')
            # title("Hello World")
            # doc += body(div('sup!'))
            # with div(cls="container"):
            #     with div(cls="row"):
            #         with div(cls="col-md-12"):
            #             with div(cls="panel panel-default"):
            #                 with div(cls="panel-heading"):
            #                     h1("Hello World")
            #                 with div(cls="panel-body"):
            #                     p("This is a paragraph")
        # print(d)

        print('------------')
        d = html()

        with d:
            with head():
                print('a!')
                p()
                print('b!')
            print('c!')
            div()
            print('d!')
            with body():
                print('e!')
                p()
                print('f!')    

            print('DONE!')
            
        # print(d)
        # d()
        # print(d())
        # with d():
        #     div()
        #     print('test')
        # print(d)
        assert str(d) == '<html><head><p></p></head><div></div><body><p></p></body></html>'


    # def test_pyml(self):
        # root = html5_parser.parse(some_html, treebuilder='dom')
        # render(root, 'test.pyml', 'pyml')


    def test_with2(self):
        d = html()
        with d:
            with head():
                with title():
                    text('Hello World')
            with body():
                with div(_class="container"):
                    with div(_class="row"):
                        with div(_class="col-md-12"):
                            with div(_class="panel panel-default"):
                                with div(_class="panel-heading"):
                                    h1('Hello World')
                                with div(_class="panel-body"):
                                    p('This is a paragraph')

        # print(d)
        print(f"{d}")
        # d()


    def test_partial(self):

        d = html()
        with d:
            with head():
                with title():
                    div(
                    '''
                    <form>
                    <label for="fname">First name:</label><br>
                    <input type="text" id="fname" name="fname" value="John"><br>
                    <label for="lname">Last name:</label><br>
                    <input type="text" id="lname" name="lname" value="Doe">
                    </form>
                    ''')
            with body():
                with div(_class="container"):
                    with div(_class="row"):
                        with div(_class="col-md-12"):
                            with div(_class="panel panel-default"):
                                with div(_class="panel-heading"):
                                    h1('Hello World')
                                with div(_class="panel-body"):
                                    p('This is a paragraph')

        print(f"{d}")


    def test_four_oh_four(self):
        # https://github.com/byteface/domonic/issues/37
        four_oh_four = html(
            head(
                meta(_charset="utf-8")
            ),
            body(
                h1("Page Not Found"),
                p("Sorry, but the page you were trying to view does not exist."),
            ),
            _lang="en"
        )
        outp = f"{four_oh_four}"
        # previous incorrect output had a closing meta tag?!. due to no __format__ rule for closed_tags
        '''
        <!DOCTYPE html>
        <html lang="en">
            <head>
                <meta charset="utf-8">
                </meta>
            </head>
            <body>
                <h1>Page Not Found</h1>
                <p>Sorry, but the page you were trying to view does not exist.</p>
            </body>
        </html>
        '''
        assert '</meta>' not in outp


    def test_boolean_attributes(self):
        # https://github.com/byteface/domonic/issues/40
        s = script(_src="foo.js", _async="async")
        print(s)
        assert '<script src="foo.js" async></script>' == str(s)

        s = script(_src="foo.js", _async='')
        print(s)
        # assert '<script src="foo.js"></script>' == str(s)

        s = script(_src="foo.js", _async='', _hidden='', _checked='')
        print(s)
        assert '<script src="foo.js" async hidden checked></script>' == str(s)

        s = script(_src="foo.js", _async=True, _hidden=False, _checked=True)
        print(s)
        assert '<script src="foo.js" async="true" hidden="false" checked="true"></script>' == str(s)


if __name__ == '__main__':
    unittest.main()



"""
    domonic.webapi.permissions
    ====================================
    https://developer.mozilla.org/en-US/docs/Web/API/Permissions
"""

class PermissionStatus:
    """
    The PermissionStatus interface represents the current state of a permission.
    """
    def __init__(self, status):
        self.status = status

    def __str__(self):
        return self.status

    def __repr__(self):
        return self.status

    @property
    def state(self):
        """
        The state of the permission.
        state replaces status
        """
        return self.status

    @state.setter
    def state(self, value):
        self.status = value


class Permissions:

    @staticmethod
    def query(PermissionDescriptor):
        """
        Returns a boolean value indicating whether the user agent has the
        specified permission.
        """
        pass

    @staticmethod
    def request(PermissionDescriptor):
        """
        Requests permission from the user agent.
        """
        pass

    # @staticmethod
    # def revoke(PermissionDescriptor):
    #     """
    #     Revokes the permission.
    #     """
    #     pass

    @staticmethod
    def requestAll():
        """
        Requests all permissions at once.
        """
        pass

    # @staticmethod
    # def revokeAll():
    #     """
    #     Revokes all permissions at once.
    #     """
    #     pass




"""
    domonic.webapi.push
    ====================================
    https://developer.mozilla.org/en-US/docs/Web/API/Push_API
"""


# class PushManager:
#     """
#     The PushManager interface provides a way for a web application to
#     register for push notifications.
#     """
#     def getSubscription(self):
#         """
#         Returns the subscription object for the current document.
#         """
#         raise NotImplementedError()

#     def permissionState(self):
#         """
#         Returns the permission state of the current document.
#         """
#         raise NotImplementedError()

#     def subscribe(self, options):
#         """
#         Subscribes the document to push notifications.
#         """
#         raise NotImplementedError()

#     def unsubscribe(self):
#         """
#         Unsubscribes the document from push notifications.
#         """
#         raise NotImplementedError()


"""
    domonic.d3.timer
    ====================================

    # TODO - completely untested

"""

from domonic.javascript import *

frame = 0  # is an animation frame pending?
timeout = None # 0  # is a timeout pending?
interval = 0  # are any timers active?
pokeDelay = 1000  # how frequently we check for clock skew
taskHead = None
taskTail = None
clockLast = 0
clockNow = 0
clockSkew = 0
clock = performance  # if isinstance(performance, object) and performance.now else Date
# setFrame = window.requestAnimationFrame.bind(window)  # if isinstance( window, object) and window.requestAnimationFrame ?  : function(f) { setTimeout(f, 17); }

def fr(f):
    """ framerate """
    setTimeout(f, 17)
setFrame = fr


def now():
    global clockNow
    global clockSkew
    clockNow = clock.now() + clockSkew
    return clockNow or (setFrame(clearNow), clockNow)

def clearNow():
    global clockNow
    clockNow = 0

class Timer():

    def __init__(self):
        self._call = None
        self._time = None
        self._next = None

    def restart(self, callback, delay, time):

        global taskTail
        global taskHead

        if not callable(callback):
            raise Error("callback is not a function")

        b = 0 if delay == None else delay
        time = now() if time == None else time + b

        if not self._next and taskTail != self:
            if taskTail:
                taskTail._next = self
        else:
            taskHead = self

        taskTail = self

        self._call = callback
        self._time = time
        sleep()

    def stop(self):
        if (self._call):
            self._call = None
            self._time = Infinity
        sleep()


def timer(callback, delay, time):
    t = Timer()
    t.restart(callback, delay, time)
    return t


def timerFlush():
    global frame
    now()  # Get the current time, if not already set.
    frame += 1  # Pretend we’ve set an alarm, if we haven’t already.
    t = taskHead
    #   e
    while t:
        e = clockNow - t._time
        if e >= 0:
            t._call.call(None, e)
        t = t._next
    frame -= 1


def wake():
    global timeout
    global clockLast
    global clockNow
    global frame
    clockLast = clock.now()
    clockNow = (clockLast) + clockSkew
    frame = timeout = 0
    try:
        timerFlush()
    finally:
        frame = 0
        nap()
        clockNow = 0


def poke():
    global clockLast
    global clockSkew
    now = clock.now()
    delay = now - clockLast
    if delay > pokeDelay:
        clockSkew -= delay
        clockLast = now


def nap():
    global taskHead
    global taskTail
    # t0
    t1 = taskHead
    # t2
    time = Infinity
    while t1:
        if (t1._call):
            if time > t1._time:
                time = t1._time
                t0 = t1
                t1 = t1._next
            else:
                t2 = t1._next
                t1._next = None
                t0._next = t2
                taskHead = t2
                t1 = t0._next if t0 else taskHead
        taskTail = t0
        sleep(time)


def sleep(time=0):
    global frame
    global clockNow
    global clockLast
    global timeout
    global interval
    if frame:
        return  # Soonest alarm already set, or will be.
    print(timeout)
    if timeout is not None:
        timeout = Global.clearTimeout(timeout)
    delay = time - clockNow  # Strictly less than if we recomputed clockNow.
    if delay > 24:
        if time < Infinity:
            timeout = Global.setTimeout(wake, time - clock.now() - clockSkew)
        if interval:
            interval = Global.clearInterval(interval)
        else:
            if not interval:
                clockLast = clock.now()
                interval = Global.setInterval(poke, pokeDelay)
        frame = 1
        setFrame(wake)

def timeout(callback, delay, time):
    t = Timer()
    delay = 0 if delay == None else delay

    def elapsed():
        t.stop()
        callback(elapsed + delay)
    t.restart(elapsed, delay, time)
    return t


def interval(callback, delay=None, time=None):

    t = Timer()
    total = delay
    if delay == None:
        t.restart(callback, delay, time)
        return t
    t._restart = t.restart

    def r(callback, delay, time):
        delay = delay
        time = now() if time == None else time

        def tick(elapsed):
            nonlocal total
            elapsed += total
            total += delay
            t._restart(tick, total, time)
            callback(elapsed)
        t._restart(tick, delay, time)

    t.restart = r
    t.restart(callback, delay, time)
    return t



"""
    domonic.webapi.XMLHttpRequest
    ====================================
    https://developer.mozilla.org/en-US/docs/Web/API/XMLHttpRequest
"""

# - TODO - create package for this and relocate FormData to it? (i think that's currently in js)
# XMLHttpRequest
# class XMLHttpRequest():

#     def __init__(self, url=None, responseType=None, withCredentials=False, timeout=0, onload=None, onerror=None, onprogress=None, ontimeout=None):
#         self.url = url
#         self.responseType = responseType
#         self.withCredentials = withCredentials
#         self.timeout = timeout
#         self.onload = onload
#         self.onerror = onerror
#         self.onprogress = onprogress
#         self.ontimeout = ontimeout
#         self.response = None
#         self.status = None


class FormData(object):

    def __init__(self, form):
        """ creates a new FormData object. """
        # TODO - parse to domonic.
        # if isinstance(form, str):
        #   self._data = domonic.loads(form) # TODO - parser wont be done enough yet
        # if isinstance(form, Node):
            # self._data = form
            # self._formobj = # TODO - serialize to object
        raise NotImplementedError

    def append(self, name, value, filename):
        """ Appends a new value onto an existing key inside a FormData object,
        or adds the key if it does not already exist. """
        raise NotImplementedError

    def delete(self, name):
        """ Deletes a key/value pair from a FormData object. """
        raise NotImplementedError

    def entries(self):
        """ Returns an iterator allowing to go through all key/value pairs contained in this object. """
        raise NotImplementedError

    def get(self, name):
        """ Returns the first value associated with a given key from within a FormData object. """
        raise NotImplementedError

    def getAll(self, name):
        """ Returns an array of all the values associated with a given key from within a FormData """
        raise NotImplementedError

    def has(self, name):
        """ Returns a boolean stating whether a FormData object contains a certain key."""
        raise NotImplementedError

    def keys(self):
        """ Returns an iterator allowing to go through all keys of the key/value pairs contained in this object."""
        raise NotImplementedError

    def set(self, name, value, filename):
        """ Sets a new value for an existing key inside a FormData object,
        or adds the key/value if it does not already exist."""
        raise NotImplementedError

    def values(self):
        """ Returns an iterator allowing to go through all values  contained in this object."""
        raise NotImplementedError
